import matplotlib.pyplot as plt
import numpy as np
import scipy.ndimage as ndimage
import dlab.psth_and_raster as par
from dlab.generalephys import placeAxesOnGrid, smoothRF
from matplotlib.patches import Rectangle, Circle
from matplotlib.collections import PatchCollection
import scipy.optimize as opt
from skimage.measure import label, regionprops
from skimage.morphology import closing
from tqdm import tqdm
import matplotlib.patches as mpatches

#smooth a 2D image, meant to be space-space of a receptive field
#size = number of pixels to smooth over
def smoothRF(img,size=3):
    """Spatially smooths a computed spatial receptive fields of an already computed receptive field. uses a 2D Gaussian to smooth.

    Parameters
    ----------
    img : np.array
        the 2d spatial receptive field to smooth
    size : int, optional
        the width of the Gaussian used to smooth (default is
        3)

    Returns
    -------
    np.array
        the smoothed input image
    """
    smooth = ndimage.gaussian_filter(img,(size,size))
    return smooth

from scipy.signal.windows import boxcar
from scipy.signal import convolve
def smooth_boxcar(data,boxcar_size=3):
    """smooths an impulse respone of an already computed receptive field. uses a boxcar to smooth.

    Parameters
    ----------
    data : np.array
        the 1d temporal kernel to smooth
    size : int, optional
        the width of the boxcar used to smooth (default is
        3)

    Returns
    -------
    np.array
        the smoothed input kernel
    """
    smoothed = convolve(data,boxcar(boxcar_size))/boxcar_size
    smoothed = smoothed[boxcar_size/2:len(data)+(boxcar_size/2)]
    return smoothed

def plotsta(sta,taus=(np.linspace(-10,280,30).astype(int)),colorrange=(-0.15,0.15),title='',taulabels=False,nrows=3,cmap=plt.cm.seismic,smooth=None,window = [[0,64],[0,64]]):
    """show the space-space plots of an already computed receptive field for a range of taus.

    Parameters
    ----------
    sta : np.array
        the 1d temporal kernel to smooth
    taus : int, optional
        the width of the boxcar used to smooth (default is
        3)
    colorrange : tuple, optional
        the limits of the colormap. default=(0.35,0.65)
    cmap :string, optional
        the color map to use to plot the receptive field. use any matplotlib colrmap. https://matplotlib.org/stable/gallery/color/colormap_reference.html   default = 'gaussian_2D'
    title : string, optional
        title of the plot. default = ''
    taulabels : bool, optional
            whether or not to label each subplot with the tau. default = False
    nrows : int, optional
        the number of rows of taus to plot. figures out how many columns to plot based on this. default = 3 
    smooth : int, optional
        the width of a Gaussian kernel for spatial smoothing. uses smoothRF function.  default = None
    window : tuple or tuple-like
        the subarea of the receptive field to plot. default = [[0,64],[0,64]]

    Returns
    -------
    np.array
        the smoothed input kernel
    """    
    ncols = np.ceil(len(taus) / nrows ).astype(int)#+1
    fig,ax = plt.subplots(nrows,ncols,figsize=(10,6))
    titleset=False
    m=np.mean(sta[str(taus[3])])
    for i,tau in enumerate(taus):
        axis = ax[int(np.floor(i/ncols))][i%ncols]
        if smooth is None:
            img = sta[str(tau)].T 
        else:
            img = smoothRF(sta[str(tau)].T,smooth)

        axis.imshow(img,cmap=cmap,vmin=colorrange[0],vmax=colorrange[1],interpolation='none')
        axis.set_frame_on(False);
        axis.set_xticklabels('',visible=False);
        axis.set_xticks([]);
        axis.set_yticklabels('',visible=False);
        axis.set_yticks([])
        axis.set_aspect(1.0)
        axis.set_xlim(window[0][0],window[0][1])
        axis.set_ylim(window[1][0],window[1][1])
        if taulabels:
            axis.set_title('tau = '+str(tau),fontsize=8)
        if titleset is not True:
            axis.set_title(title,fontsize=12)
            titleset=True
        else:
            if tau == 0:
                axis.set_title('tau = '+str(tau),fontsize=12)
    plt.tight_layout()
    fig.show()

def twoD_Gaussian(p, amplitude, xo, yo, sigma_x, sigma_y, theta, offset):
    """function for fitting a spatial receptive with a 2-dimensional gaussian.

    Parameters
    ----------
    p : tuple
        the center of the fit
    amplitude : float
        the scaling of the gaussian
    xo : float
        starting x position of fit
    yo : float
        starting y position of fit
    sigma_x : float
        the width of the gaussian in the x dimension
    sigma_y : float
        the width of the gaussian in the y dimension
    theta : float
        the angle between the major and minor axes of the gaussian
    offset : float
        a fixed offset from baseline 

    Returns
    -------
    np.array
        fit 
    """
    x=p[0];y=p[1]
    xo = float(xo)
    yo = float(yo)    
    a = (np.cos(theta)**2)/(2*sigma_x**2) + (np.sin(theta)**2)/(2*sigma_y**2)
    b = -(np.sin(2*theta))/(4*sigma_x**2) + (np.sin(2*theta))/(4*sigma_y**2)
    c = (np.sin(theta)**2)/(2*sigma_x**2) + (np.cos(theta)**2)/(2*sigma_y**2)
    g = offset + amplitude*np.exp( - (a*((x-xo)**2) + 2*b*(x-xo)*(y-yo) 
                            + c*((y-yo)**2)))
    return g.ravel()

def fit_rf_2Dgauss(data,center_guess,width_guess=2,height_guess=2):
    """function for fitting a spatial receptive with a 2-dimensional gaussian.

    Parameters
    ----------
    data : tuple
        the data to be fit
    center_guess : tuple, ints
        the center of the fit
    width_guess : int, optional
        the x width position of fit
    height_guess : int, optional 
        the y height  of fit

    Returns
    -------
    popt:  np.array
        fit parameters (popt from opt.curve_fit)
    pcov: np.array
        covariance of parameter (pcov from opt.curve_fit)
    reshaped_to_space: np.array
        the fit reshaped to the dimensions of the input array 
    """
    dataToFit = data.ravel()
    x=np.linspace(0,np.shape(data)[0]-1,np.shape(data)[0])
    y=np.linspace(0,np.shape(data)[1]-1,np.shape(data)[1])
    x, y = np.meshgrid(x, y)
    popt,pcov = opt.curve_fit(twoD_Gaussian,(x,y),dataToFit,p0=(data[center_guess[1]][center_guess[0]], center_guess[1], center_guess[0], width_guess, height_guess, 0, 0))
    reshaped_to_space=(x,y,twoD_Gaussian((x,y),*popt).reshape(np.shape(data)[1],np.shape(data)[0]))
    return popt,pcov,reshaped_to_space

def fit_rf_2Dgauss_centerFixed(data,center_guess,width_guess=2,height_guess=2):
    """function for fitting a spatial receptive with a 2-dimensional gaussian. similar to fit_rf_2DGauss, but the position of the Gaussian cannot change with the fit, only the other parameters

    Parameters
    ----------
    data : tuple
        the data to be fit
    center_guess : tuple, ints
        the center of the fit. this will not change during fitting
    width_guess : int, optional
        the x width position of fit
    height_guess : int, optional 
        the y height  of fit

    Returns
    -------
    np.array,np.array,np.array
         fit parameters (popt from opt.curve_fit), covariance of parameter (pcov from opt.curve_fit), the fit reshaped to the dimensions of the input array 
    """
    dataToFit = data.ravel()
    x=np.linspace(0,np.shape(data)[0]-1,np.shape(data)[0])
    y=np.linspace(0,np.shape(data)[1]-1,np.shape(data)[1])
    x, y = np.meshgrid(x, y)
    
    def twoD_Gaussian_fixed(p, amplitude,sigma_x, sigma_y, theta, offset):
        x=p[0];y=p[1]
        a = (np.cos(theta)**2)/(2*sigma_x**2) + (np.sin(theta)**2)/(2*sigma_y**2)
        b = -(np.sin(2*theta))/(4*sigma_x**2) + (np.sin(2*theta))/(4*sigma_y**2)
        c = (np.sin(theta)**2)/(2*sigma_x**2) + (np.cos(theta)**2)/(2*sigma_y**2)
        g = offset + amplitude*np.exp( - (a*((x-center_guess[1])**2) + 2*b*(x-center_guess[1])*(y-center_guess[0]) 
                                + c*((y-center_guess[0])**2)))
        return g.ravel()
    popt,pcov = opt.curve_fit(twoD_Gaussian_fixed,(x,y),dataToFit,p0=(data[int(center_guess[0])][int(center_guess[1])], width_guess, height_guess, 0, 0))
    reshaped_to_space=(x,y,twoD_Gaussian_fixed((x,y),*popt).reshape(np.shape(data)[1],np.shape(data)[0]))
    return popt,pcov,reshaped_to_space

def impulse(sta,center,taus = np.arange(-10,290,10).astype(int)):
    """find the impulse response at a certain position across a range of taus for an already computed receptive field

    Parameters
    ----------
    sta : tuple
        the data to be fit
    center : tuple, ints
        the x,y position to meausure the impulse response
    taus : np.array, optional
        the interval (ints) at which the measure the impulse response. should match the taus of the linear receptive field 

    Returns
    -------
    taus: np.array
        the interval (times) of the impulse response
    impulse: np.array
        the impulse response
    """
    impulse = [sta[str(tau)][int(center[0])][int(center[1])] for tau in taus]
    return (taus,impulse)

#is very, very finicky right now, requires lots of manual tweaking. 
def fitRF(RF,threshold=None,fit_type='gaussian_2D',verbose=False,rfsizeguess=1.2,flipSpace=False,backup_center=None,zoom_int=10,zoom_order=5,centerfixed=False):
    """find the impulse response at a certain position across a range of taus for an already computed receptive field

        Parameters
        ----------
        RF : tuple
            the data to be fit
        threshold : tuple, optional
            the x,y position to meausure the impulse response
        fit_type :string, optional
            the type of fit to try. options are 'gaussian_2D', . default = 'gaussian_2D'
        verbose : bool, optional
            print extra statements during the fit process. default = False
        rfsizeguess : float, optional
            default = 1.2
        flipSpace : bool, optional
            whether or not to transpose the axes. default = False
        backup_center : tuple, optional
            a backup center to use to zoom around if the fit doesn't work. default = None
        zoom_int : int, optional
            the zoom factor of the zoomed in receptive field around the center. default = 10. 
        zoom_order : int, optional
            the order of the filter to smooth a zoomed in receptive field. default = 5
        centerfixed : bool, optional
            whether or not to use a fixed center for the fit. default = False
            
        Returns
        -------
        fit: dictionary
            the averaged spatial RF, 
            the centroid of the fit [max fit]
            a 2D gaussian fit of that spatial RF
            the impulse response at the center of the fit
        """
#   TODO: a fit of that impulse response with: ?? currently not defined.
    if False:#if np.isnan(RF[RF.keys()[0]][0][0]):#check to make sure there is any data in the STRF to try to fit. if not, return the correct data structure filled with None
        fit={};fit['avg_space_fit']=None;fit['params'] = None;fit['cov']=None ;fit['amplitude']=None ;fit['x']=None ;fit['y']=None ;fit['s_x']=None ;fit['s_y']=None ;fit['theta']=None ;fit['offset']=None;fit['center']=None;fit['peakTau']=None;fit['impulse']=None;fit['roughquality']=None
        return fit
    else:

        
        if type(RF)==dict:
            if 'fit' in RF.keys():
                trash = RF.pop('fit') # get rid of the dictionary entry 'fit'; we only want sta tau frames in this dictionary
                taus = [int(i) for i in RF.keys()]
                taus.sort()
                fit={}
            #========================================================================
            #find the taus to average over for the spatial RF
            #first, define the threshold; above this means there is a non-noise pixel somwhere in the space-space
            if threshold == None:
                #set to z sd above mean
                blank = (RF['-10']+RF['0']+RF['10'])/3. # define a blank, noise-only image
                threshold = np.mean(blank)+np.std(blank)*3.
                if verbose:
                    print('threshold: '+str(threshold))
                
            #find the average space-space over only the range of non-noise good 
            avgRF = np.zeros(np.shape(RF[str(int(taus[0]))]))#initialize the average to blank.
            goodTaus = [40,50,60,70,80,90,100]#
            for tau in goodTaus:
                avgRF += RF[str(int(tau))]
            avgRF = avgRF / float(len(goodTaus))
            fit['avg_space']=avgRF

            maximum_deviation = 0;best_center = (0,0)
            for i in np.linspace(24,63,40):
                for j in np.linspace(10,49,40):
                    imp_temp = impulse(RF,(i,j))
                    if np.max(np.abs(imp_temp[1])) > maximum_deviation:
                        best_center = (int(i),int(j))
                        maximum_deviation = np.max(np.abs(imp_temp[1]))
            center = best_center
            imp_temp = impulse(RF,center)
            if verbose:
                print('peak frame tau: '+str(int(imp_temp[0][np.where(np.array(np.abs(imp_temp[1]))==np.max(np.abs(imp_temp[1])))[0][0]])))
                print('peak center   : '+str(center))
                print('peak value    : '+str(RF[str(int(imp_temp[0][np.where(np.array(np.abs(imp_temp[1]))==np.max(np.abs(imp_temp[1])))[0][0]]))][center[0],center[1]]))
            peak_frame = RF[str(int(imp_temp[0][np.where(np.array(np.abs(imp_temp[1]))==np.max(np.abs(imp_temp[1])))[0][0]]))]
            peak = peak_frame[int(center[0]),int(center[1])]
            #center = (np.where(np.abs(smoothRF(peak_frame,1)) == np.max(np.abs(smoothRF(peak_frame,1))))[0][0],np.where(np.abs(smoothRF(peak_frame,1)) == np.max(np.abs(smoothRF(peak_frame,1))))[1][0])
            
            #========================================================================   
        if type(RF)==np.ndarray:
            fit={}
            fit['avg_space']=RF
            peak_frame = RF
            center = backup_center
            peak = np.max(peak_frame)
        #====fit==================================================================

        if verbose:
            print('peak amp: '+str(peak)+'  threshold: '+str(threshold))
        if np.abs(peak) > threshold * 1.0:
            peak_frame = smoothRF(ndimage.zoom(peak_frame,zoom_int,order=zoom_order),0)
            fit['roughquality']='good'
        else:
            center = backup_center
            imp_temp = impulse(RF,center)
            peak_frame = RF[str(int(100))]
            peak = peak_frame[center[0],center[1]]
            peak_frame = smoothRF(zoom(peak_frame,zoom_int,order=zoom_order),0)
            print('could not find a peak in the RF, using center: '+str(center))
            print('peak amplitude: '+str(peak)+', threshold: '+str(threshold))
            fit['roughquality']='bad'
        fit['center']=center
        fit['center_guess']=center
        fit['fit_image']=peak_frame
        if verbose:
            print('center guess: '+str(center))
        
        #initialize some empty parameters
        fitsuccess=False;retry_fit = False
        best_fit = 10000000#initialize impossibly high
        fit['avg_space_fit']=None;fit['params']=None
        best_fit_output = ((None,None,None,None,None,None,None),600,None)
        try:
            if centerfixed:
                popt,pcov,space_fit = fit_rf_2Dgauss_centerFixed(peak_frame,(center[0]*zoom_int,center[1]*zoom_int),width_guess=rfsizeguess*zoom_int,height_guess=rfsizeguess*zoom_int)
            else:
                popt,pcov,space_fit = fit_rf_2Dgauss(peak_frame,(center[0]*zoom_int,center[1]*zoom_int),width_guess=rfsizeguess*zoom_int,height_guess=rfsizeguess*zoom_int)
        except:
            popt,pcov,space_fit=((None,0,0,0,0,0,0),600,np.zeros((64,64)));print('2D fit failed')
        
        fit['avg_space_fit']=np.array(space_fit)/float(zoom_int)
        fit['params'] = popt    
        fit['cov']=pcov
        fit['amplitude']=popt[0]
        if centerfixed:
            fit['x']=center[1]
            fit['y']=center[0]
            fit['s_x']=popt[1] / float(zoom_int)
            fit['s_y']=popt[2] / float(zoom_int)
            fit['theta']=popt[3]
            fit['offset']=popt[4] / float(zoom_int)
        else:
            fit['x']=popt[1] / float(zoom_int)
            fit['y']=popt[2] / float(zoom_int)
            fit['s_x']=popt[3] / float(zoom_int)
            fit['s_y']=popt[4] / float(zoom_int)
            fit['theta']=popt[5]
            fit['offset']=popt[6] / float(zoom_int)
        #======================================================================== 
    
    #        
        #============get impulse======================================================================== 
        if verbose:
            print('center: '+str(center[0])+' '+str(center[1]))
            
        if fit['avg_space_fit'] is not None:
            center_h = (np.ceil(fit['y']),np.ceil(fit['x']))
            center_r = (np.round(fit['y']),np.round(fit['x']))
            center_l = (np.floor(fit['y']),np.floor(fit['x']))
        try:
            impuls_h = impulse(RF,center_h,taus)[1]
            impuls_r = impulse(RF,center_r,taus)[1]
            impuls_l = impulse(RF,center_l,taus)[1]
            if np.max(np.abs(impuls_h)) > np.max(np.abs(impuls_r)):
                if np.max(np.abs(impuls_h)) > np.max(np.abs(impuls_l)):
                    impuls = impuls_h
                    center = center_h
                else:
                    impuls = impuls_l
                    center= center_l
            else:
                if np.max(np.abs(impuls_r)) > np.max(np.abs(impuls_l)):
                    impuls= impuls_r
                    center= center_r
                else:
                    impuls= impuls_l
                    center= center_l
        except:
            impuls = np.array([0,0,0,0,0,0]) #impuls = np.zeros(len(taus))
            taus = impuls
        if fit_type == 'gaussian_2D':
            #get impulse at the 'center'
            if verbose:
                print('center from fit: '+str(center[0])+' '+str(center[1]))
            #impuls = [RF[str(tau)][center[0]][center[1]] for tau in taus]
            fit['impulse']=(np.array(taus),np.array(impuls))
            peakTau = taus[np.abs(np.array(impuls)).argmax()]
            peakTau = 80
            fit['peakTau']=peakTau
        
        fit['center_usedforfit'] = fit['center']
        fit['center_usedforimp'] = center
        fit['impulse']=(np.array(taus),np.array(impuls))
        #======================================================================== 
        
        return fit 

def show_sta_fit(fit,colorrange=(0.35,0.65),cmap=plt.cm.seismic,title='',contour_levels=3):
    """convenience plotting method for showing spatial and temporal filters pulled from fitting an already computed STRF

        Parameters
        ----------
        fit : dictionary
            the fit data to plot. requires 'impulse' and 'avg_space' keys. 
        colorrange : tuple, optional
            the limits of the colormap. default=(0.35,0.65)
        cmap :string, optional
            the color map to use to plot the receptive field. use any matplotlib colrmap. https://matplotlib.org/stable/gallery/color/colormap_reference.html   default = 'gaussian_2D'
        title : string, optional
            title of the plot. default = ''
        contour_levels : int, optional
            default = 3
        
            
        Returns
        -------
        fig: matplotlib figure handle
            matplotlib figure handle
        """
    if fit is not None:
        fig = plt.figure(figsize=(6,2.75))
        ax_full_space = placeAxesOnGrid(fig,dim=(1,1),xspan=(0,0.32),yspan=(0,0.5))
        ax_zoom_space = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.33,0.63),yspan=(0,0.5))
        ax_zoom_space_filtered = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.56,0.86),yspan=(0,0.5))
        ax_impulse = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.1,0.7),yspan=(0.54,1.0))
        
        ax_impulse.plot(fit['impulse'][0],np.zeros(len(fit['impulse'][0]))+0.5,'r-')
        ax_impulse.plot(fit['impulse'][0],fit['impulse'][1]);
        ax_impulse.set_ylim(colorrange[0],colorrange[1]) 
        ax_impulse.set_xlabel('time (msec)') 
        ax_impulse.set_ylabel('spike probability')
        #ax_impulse.text(1,1,str(np.mean(fit['cov'])))
        
        ax_full_space.plot([fit['y']],[fit['x']],'k+',markersize=6.0)#([fit['center'][0]],[fit['center'][1]],'k+',markersize=6.0)
        ax_full_space.imshow(fit['avg_space'].T,
           cmap=cmap,
           vmin=colorrange[0],vmax=colorrange[1],
            interpolation='none');
        plt.setp(ax_full_space.get_xticklabels(),visible=False);ax_zoom_space_filtered.xaxis.set_major_locator(plt.NullLocator())
        plt.setp(ax_full_space.get_yticklabels(),visible=False);ax_zoom_space_filtered.yaxis.set_major_locator(plt.NullLocator())
        ax_full_space.set_title(title)
        ax_full_space.set_ylim(20,40)
        ax_full_space.set_xlim(12,52)
        ax_full_space.axvline(x=32,linewidth=1,color='r',ls='--')
        
        zoom_size = 5 #number of pixels in each direction from peak center to keep in zooms
        ax_zoom_space.imshow(fit['avg_space'],
                   cmap=cmap,
                   vmin=colorrange[0],vmax=colorrange[1],
                    interpolation='none');
        #ax_zoom_space.plot([fit['x']],[fit['y']],'k+',markersize=6.0)
        ax_zoom_space.set_xlim(fit['center'][1]-zoom_size,fit['center'][1]+zoom_size)
        ax_zoom_space.set_ylim(fit['center'][0]-zoom_size,fit['center'][0]+zoom_size) 
        plt.setp(ax_zoom_space.get_xticklabels(),visible=False);ax_zoom_space_filtered.xaxis.set_major_locator(plt.NullLocator())
        plt.setp(ax_zoom_space.get_yticklabels(),visible=False);ax_zoom_space_filtered.yaxis.set_major_locator(plt.NullLocator())
        

        ax_zoom_space_filtered.imshow(smoothRF(fit['avg_space'],1),
                   cmap=cmap,
                   vmin=colorrange[0]/3.,vmax=colorrange[1]/3.,
                    interpolation='none');
        
        #ax_zoom_space_filtered.plot([fit['x']],[fit['y']],'k+',markersize=6.0)
        ax_zoom_space_filtered.set_xlim(fit['center'][1]-zoom_size,fit['center'][1]+zoom_size)
        ax_zoom_space_filtered.set_ylim(fit['center'][0]-zoom_size,fit['center'][0]+zoom_size)
        plt.setp(ax_zoom_space_filtered.get_xticklabels(),visible=False);ax_zoom_space_filtered.xaxis.set_major_locator(plt.NullLocator())
        plt.setp(ax_zoom_space_filtered.get_yticklabels(),visible=False);ax_zoom_space_filtered.yaxis.set_major_locator(plt.NullLocator())
        
    
        

        
        if 'avg_space_fit' in fit.keys():
            if fit['avg_space_fit'] is not None:
                if np.max(np.abs(fit['impulse'][1])) > np.std(fit['impulse'][1])*2.:
                    ax_full_space.add_patch(Rectangle((fit['y']-zoom_size,fit['x']-zoom_size),zoom_size*2,zoom_size*2,fill=None,ls='dotted'))
                    ax_zoom_space.contour(fit['avg_space_fit'][0],
                                fit['avg_space_fit'][1],
                                fit['avg_space_fit'][2],
                                contour_levels)
                    ax_zoom_space_filtered.contour(fit['avg_space_fit'][0],
                                fit['avg_space_fit'][1],
                                fit['avg_space_fit'][2],
                                contour_levels)
                    ax_full_space.contour(fit['avg_space_fit'][1],
                                fit['avg_space_fit'][0],
                                fit['avg_space_fit'][2],
                                contour_levels)
                    ax_full_space.set_aspect('equal')   

		#                   vmin=-0.08,vmax=0.08);
        #plt.tight_layout()
        return fig
    
#TODO: should be modified to take data of any shape (for example, an LFP trace) to average.
def sta(spiketimes,data,datatimes,taus=(np.linspace(-10,280,30)),exclusion=None,samplingRateInkHz=25):
    """compute a spike-triggered average on three dimensional data. this is typically a movie of the stimulus

        Parameters
        ----------
        spiketimes : np.array
            the times of spikes
        data : tuple-like or np.array
            the stimulus data, with one dimension matching datatimes
        datatimes np.array
            the times of each stimulus in data, should match one of the axes of data
        taus : tuple of np.array, optional
            the taus to calculate. usually provided in milliseconds. default = (np.linspace(-10,280,30))
        exclusion : int, optional
            default = 3
        samplingRateInkHz: float, optional
            multiply the taus by this number.  default=25
        
            
        Returns
        -------
        output: dict
            the average of the input at each tau, where the tau is the key and the computed average is the value. 
        """
    output = {}
    for tau in taus:
        avg = np.zeros(np.shape(data[:,:,0]))
        count = 0
        for spiketime in spiketimes:
            if spiketime > datatimes[0] and spiketime < datatimes[-1]-0.6:
                if exclusion is not None: #check to see if there is a period we are supposed to be ignoring, because of eye closing or otherwise
                    if spiketime > datatimes[0] and spiketime < datatimes[-1]-0.6:
                        index = (np.where(datatimes > (spiketime - tau*samplingRateInkHz))[0][0]-1) % np.shape(data)[2]
                        avg += data[:,:,index]
                else:
                    index = (np.where(datatimes > (spiketime - tau*samplingRateInkHz))[0][0]-1) % np.shape(data)[2]
                    avg += data[:,:,index]
                count+=1
        output[str(int(tau))]=avg/count
    return output

#should compute a spike-triggered average on n-dimensional data.
#this is typically a movie of the stimulus of shape [frames,x,y], but could be any shape, such as a 1-d noise stimulus or continuous trace.
#as of now, limited to computation of the average along the first axis of 'data'. 
def sta2(spiketimes,data,datatimes,taus=(np.linspace(-10,280,30)),exclusion=None,samplingRateInkHz=30,time_domain=False):
    """#should compute a spike-triggered average on n-dimensional data. this is typically a movie of the stimulus of shape [frames,x,y], but could be any shape, such as a 1-d noise stimulus or continuous trace. as of now, limited to computation of the average along the first axis of 'data'. 

        Parameters
        ----------
        spiketimes : dictionary
            the fit data to plot. requires 'impulse' and 'avg_space' keys. 
        data : tuple, optional
            the limits of the colormap. default=(0.35,0.65)
        datatimes :string, optional
            the color map to use to plot the receptive field. use any matplotlib colrmap. https://matplotlib.org/stable/gallery/color/colormap_reference.html   default = 'gaussian_2D'
        taus : tuple of np.array, optional
            the taus to calculate. usually provided in milliseconds. default = (np.linspace(-10,280,30))
        exclusion : int, optional
            default = 3
        samplingRateInkHz: float, optional
            multiply the taus by this number.  default=30
        time_domain: bool, optional
              default=False

        Returns
        -------
        output: dict
            the average of the input at each tau, where the tau is the key and the computed average is the value. 
        """
    output = {}
    if time_domain:
        taus = taus / 1000.
    else:
        taus = (taus * samplingRateInkHz)/1000.
		
    for tau in taus:
        avg = np.zeros(np.shape(data[0]))
        count = 0
        for spiketime in spiketimes:
            if spiketime > datatimes[0] and spiketime < datatimes[-1]-0.5:
                if exclusion is not None: #check to see if there is a period we are supposed to be ignoring, because of eye closing or otherwise
                    if spiketime > datatimes[0] and spiketime < datatimes[-1]-0.5:
                        index = (np.where(datatimes > (spiketime - tau))[0][0]-1) % np.shape(data)[0]
                        avg += data[index]
                else:
                    index = (np.where(datatimes > (spiketime - tau))[0][0]-1) % np.shape(data)[0]
                    avg += data[index]
                count+=1
        output[str(tau)]=avg/count
    return output

class rCorr():
    """
    Class for reverse-correlation analyses such as STA, STC, etc. All analyses take spike times, stimulus times, signal, and tau values. Would like to accommodate functions for 1D and 2D signals. 
    """
    def __init__(self,spike_times,stim_times,signal,taus=np.linspace(-0.01,0.28,30),exclusion=None):
        self.signal      = np.array(signal)
        self.stim_times  = np.array(stim_times)
        self.spike_times = np.array(spike_times)
        self.taus        = np.array(taus)
        
        self.start       = stim_times[0]
        self.end         = stim_times[-1]-0.06
        
        self.stim_spikes = self.spike_times[(self.spike_times >= self.start) & (self.spike_times < self.end)]
        self.spikes_adj  = self.stim_spikes[:,np.newaxis] - self.taus
        
        self.spikes_adj  = self.spikes_adj.T
        
        def find_nearest(array, value):
            idx = (np.abs(array-value)).argmin()
            return idx
        
        if exclusion is not None: 
            # Check if there are time periods we should ignore (eye closing, stim issues, etc.)
            for i in exclusion:
                ex1 = find_nearest(self.stim_spikes,i[0])
                ex2 = find_nearest(self.stim_spikes,i[1])
                self.stim_spikes = np.delete(self.stim_spikes,np.arange(ex1,ex2))
    
    def sta(self):
        output = np.zeros(((len(self.taus),) + self.signal[0].shape))
        output[:] = np.nan
        
        for i,tau in enumerate(self.taus):
            avg = np.zeros(self.signal[0].shape)
            count = 0
            for spike in self.stim_spikes:
                index = (np.where(self.stim_times > (spike - tau))[0][0]-1)
                avg += self.signal[index]
                count+=1
                    
            output[i,:,:] = avg/count
            
        return output,self.taus
        
    def stc(self):
        return 

def plot_sta(sta,taus=np.linspace(-0.01,0.28,30),nrows=3,smooth=None,taulabels=False,**kwargs):
    img = sta
    
    if smooth is not None:
        for i,rf in enumerate(img):
            img[i] = smoothRF(rf,size=smooth)
            
    img = (img - img.mean())/img.std()

    # gmin = img.mean()-(img.std()*3)
    # gmax = img.mean()+(img.std()*3)
    gmin = -4.5
    gmax = 4.5
    
    ncols  = np.ceil(len(taus)/nrows).astype(int)
    
    if 'cmap' in kwargs:
        colormap = kwargs['cmap']
    else: colormap='bwr'
    
    with mpl.rc_context({'xtick.bottom':False,
                          'xtick.labelbottom':False,
                          'ytick.left':False,
                          'ytick.labelleft':False
                          }):
        fig,ax = plt.subplots(nrows,ncols,figsize=(10,6))
        
        for i,tau in enumerate(taus):
            axis = ax[int(np.floor(i/ncols))][i%ncols]
            axis.imshow(np.fliplr(img[i]),vmin=gmin,vmax=gmax,cmap=colormap)
            axis.set_frame_on(False)
            axis.set_aspect(1.0)
            if taulabels==True:
                axis.set_title(f'{int(tau*1000)}ms',fontsize=8,color='k')
            else:
                if tau==0:
                    axis.set_title('0ms',fontsize=8,color='k')
                
    if 'title' in kwargs:
        fig.suptitle(kwargs['title'] +f' min={img.min():.3f} max={img.max():.3f}',
                     fontsize=10,color='k',y=0.85,fontweight='semibold')
        
    plt.subplots_adjust(wspace=0.1,hspace=-0.6)
    
    if 'facecolor' in kwargs:
        fig.set_facecolor(kwargs['facecolor'])
        

def sta_with_subfields(spiketimes,data,datatimes,taus=(np.linspace(-10,280,30)),exclusion=None,samplingRateInkHz=25):
    """compute a spike-triggered average on three dimensional data. this is typically a movie of the stimulus, similar to sta(). the difference is that this calculates bright and dark separately as well.

        Parameters
        ----------
        spiketimes : np.array
            the times of spikes
        data : tuple-like or np.array
            the stimulus data, with one dimension matching datatimes
        datatimes np.array
            the times of each stimulus in data, should match one of the axes of data
        taus : tuple of np.array, optional
            the taus to calculate. usually provided in milliseconds. default = (np.linspace(-10,280,30))
        exclusion : int, optional
            default = 3
        samplingRateInkHz: float, optional
            multiply the taus by this number.  default=25
        
            
        Returns
        -------
        output: dict
            the average of the input at each tau, where the tau is the key and the computed average is the value. 
        output:_ON dict
            the average of the input bright pixels of the input only at each tau, where the tau is the key and the computed average is the value. 
        output_OFF: dict
            the average of the input dark pizxels of the input only at each tau, where the tau is the key and the computed average is the value. 
        """    
    output = {}
    output_ON = {}
    output_OFF = {}
    for tau in taus:
        avg = np.zeros(np.shape(data[:,:,0]))
        avg_ON = np.zeros(np.shape(data[:,:,0]))
        avg_OFF = np.zeros(np.shape(data[:,:,0]))
        count = 0
        for spiketime in spiketimes:
            if spiketime > datatimes[0] and spiketime < datatimes[-1]-0.6:
                if exclusion is not None: #check to see if there is a period we are supposed to be ignoring, because of eye closing or otherwise
                    if spiketime > datatimes[0] and spiketime < datatimes[-1]-0.6:
                        index = (np.where(datatimes > (spiketime - tau*samplingRateInkHz))[0][0]-1) % np.shape(data)[2]
                        avg += data[:,:,index]
                        avg_ON += (data[:,:,index] + 1 ) / 2
                        avg_OFF += (data[:,:,index]- 1 ) / 2
                else:
                    index = (np.where(datatimes > (spiketime - tau*samplingRateInkHz))[0][0]-1) % np.shape(data)[2]
                    avg += data[:,:,index];avg_ON += (data[:,:,index] + 1 ) / 2;avg_OFF += (data[:,:,index] - 1 ) / 2
                count+=1
        output[str(int(tau))]=avg/count
        output_ON[str(int(tau))]=avg_ON/count# - 0.5
        output_OFF[str(int(tau))]=avg_OFF/count# + 0.5
    return output,output_ON,output_OFF

#from https://github.com/JesseLivezey/gabor_fit/blob/master/gabor_fit/fit.py
def generate_gabor(p, pixels_x,pixels_y, theta, stdx, stdy, lamb, phase):
    """
    Generate a gabor filter based on given parameters.
    Parameters
    ----------
    pixels : tuple of ints
        Height and width of patch.
    p : (x,y)
        x : float
            x Location of center of gabor in pixels.
        y : float
            y Location of center of gabor in pixels.
    pixels_x: float
        number of pixels in x dimension of the fit
    pixels_y: float
        number of pixels in y dimension of the fit
    theta : float
        Rotation of gabor in plane in degrees.
    stdx : float
        Width of gaussian window along rot(x) in pixels.
    stdy : float
        Width of gaussian window along rot(y) in pixels.
    lamb : float
        Wavelength of sine funtion in pixels along rot(x).
    phase : float
        Phase of sine function in degrees.
    Returns
    -------
    gabor : ndarray
        normalized fit parameters
    """
    #print (pixels_x,pixels_y)
    x=p[0];y=p[1]
    x_coords = np.arange(0, int(pixels_x))#pixels[0])
    y_coords = np.arange(0, int(pixels_y))#pixels[1])
    xx, yy = np.meshgrid(x_coords, y_coords)
    unit2rad = 2. * np.pi 
    deg2rad = 2. * np.pi / 360.
    xp = (xx - x) * np.cos(deg2rad * theta) - (yy - y) * np.sin(deg2rad * theta)
    yp = (xx - x) * np.sin(deg2rad * theta) - (yy - y) * np.cos(deg2rad * theta)
    gabor = (np.exp(-xp**2 / (2. * stdx**2) - yp**2 / (2. * stdy**2)) *
             np.sin(unit2rad * (xp / lamb) - deg2rad * phase))
    norm = np.sqrt((gabor**2).sum())
    g = gabor/norm
    return g.ravel()

def fitgabor_2(data,pixels=(10,10),x=32.,y=32.,theta=0.,stdx=3.,stdy=3.,lamb=1.5,phase=0.):
    """function for fitting a spatial receptive with a 2-dimensional Gabor.

    
    Parameters
    ----------
    data : tuple
        the data to be fit
    pixels : tuple, ints
        the number of pixels in the fit
    x: float
        starting x position of fit
    y : float
        starting y position of fit
    height_guess : int, optional 
        the y height  of fit
    theta : float
        the angle between the major and minor axes of the gaussian
    stdx : float
        the width of the gaussian in the x dimension
    stdy : float
        the width of the gaussian in the y dimension
    lamb : float
        Wavelength of sine funtion in pixels along rot(x).
    phase : float
        Phase of sine function in degrees.

    Returns
    -------
    popt:  np.array
        fit parameters (popt from opt.curve_fit)
    pcov: np.array
        covariance of parameter (pcov from opt.curve_fit)
    reshaped_to_space: np.array
        the fit reshaped to the dimensions of the input array 
    """
    popt,pcov = opt.curve_fit(generate_gabor,(x,y),data.ravel(),p0=(int(pixels[0]),int(pixels[1]),theta,stdx,stdy,lamb,phase))
    reshaped_to_space=(x,y,generate_gabor((x,y),*popt).reshape(np.shape(data)[1],np.shape(data)[0]))
    return popt,reshaped_to_space,pcov

def check_rfs_in_df(df_rf,sds=4):
    """this function will check if the RFs computed and stored in this dataframe have signal in them, adding three rows to the input df:
          rf_computed: if any RF STA was calculated at all
          good: if there was a non-noise RF
          rf_color: the color of the noise used to generate the "good" RF, if there was a good one

    
    Parameters
    ----------
    df_rf:  pandas.DataFrame
        a pandas DataFrame with some receptive fields computed from uv and green noise to check. must have a 'g_avg_space' column, and 'g_fit_image' and 'u_fit_image' columns
    sds : int, optional
        the number of SDs a pixels has to be above in order to be considered a real RF pixel

    Returns
    -------
    df_rf:  pandas.DataFrame
        the input DataFrame 
    """
    x=sds
    rf_computed = []
    good_rf = []
    rf_color = []
    for ind in df_rf.index:
        if type(df_rf.g_avg_space[ind]) != str:
            # measure some things about the computed rf
            rf_computed.extend([True])
            
            gr = df_rf.g_fit_image[ind]
            im = smoothRF(gr,.5);
            thresh = np.mean(im)+x*np.std(im)
            im[np.abs(smoothRF(gr,.6))<thresh]=0
            gr_pixels = np.nansum(np.abs(im))
            
            uv = df_rf.u_fit_image[ind]
            im = smoothRF(uv,.5);
            thresh = np.mean(im)+x*np.std(im)
            im[np.abs(smoothRF(uv,.6))<thresh]=0
            uv_pixels = np.nansum(np.abs(im))
    
            if gr_pixels>0 or uv_pixels>0:
                good_rf.extend([True])
                if uv_pixels>gr_pixels:
                    im = smoothRF(uv,.5)
                    rf_color.extend(['uv'])
                else:
                    im = smoothRF(gr,.5)
                    rf_color.extend(['green'])
            else: 
                rf_color.extend([np.nan]);good_rf.extend([False])
        else:
            rf_computed.extend([False])
            good_rf.extend([False])
            rf_color.extend([np.nan])
    df_rf['rf_computed']=rf_computed
    df_rf['good_rf']=good_rf
    df_rf['rf_color']=rf_color
    
    return df_rf

def show_impulse(a,center):
    """this function compute an impulse response from a receptive field and plot it. both the impulse and the spatial receptive field at 80 msec tau are plotted.  

    
    Parameters
    ----------
    a:  dict
        a dictionary containing a computed receptive field, where the keys of the dict are the tau and the values are the spatial RF at that tau
    center : tuple, ints
        the pixels in each receptive field at which to measure the impulse response

    Returns
    -------
    None
    """

    center=(41,32)
    i = impulse(a,center,taus=np.linspace(-10,280,30).astype(int))
    plt.plot(i[0],i[1])
    plt.ylim(108,148)
    plt.gca().axhline(128,ls='--')
    plt.figure()
    plt.imshow(a['80'])
    plt.gca().scatter(center[1],center[0],color='w',alpha=0.3)

def segRF(array, flip=False, kernel=1,colormap='PiYG'):
    '''
    Highlights RF using thresholding and image segmentation. Requires manual selection of the index (tau value) of interest in the series of RFs
    '''
    im = array
    img_gauss = smoothRF(im,kernel)

    bn_img = np.zeros([img_gauss.shape[0],img_gauss.shape[1]])
    bn_img = abs(img_gauss) >= (abs(img_gauss).max()-np.std(img_gauss))

    sklabels = label(closing(bn_img))

    sizes = []
    for region in regionprops(sklabels):
        sizes.append(region.area)

    max_roi = np.asarray(sizes).argmax()


    centroid = regionprops(sklabels)[max_roi].centroid
    centroid = (centroid[1],centroid[0])
    major_length = regionprops(sklabels)[max_roi].major_axis_length
    minor_length = regionprops(sklabels)[max_roi].minor_axis_length
    eccentricity = regionprops(sklabels)[max_roi].eccentricity
    orientation = regionprops(sklabels)[max_roi].orientation*180/np.pi
    if flip == True:
        orientation = -orientation
    perimeter = regionprops(sklabels)[max_roi].perimeter
    
    data = {'centroid':centroid,
            'major_length':major_length,
            'minor_length':minor_length,
            'eccentricity': eccentricity,
            'orientation': orientation,
            'perimeter':perimeter
           }

    ellipse = mpatches.Ellipse(centroid,minor_length,major_length,orientation,
                              fill=False, edgecolor='r', linewidth=2)
    
    fig,ax = plt.subplots()
    ax.imshow(img_gauss,cmap=colormap,vmin=img_gauss.min(),
            vmax=(abs(img_gauss).max()-np.std(img_gauss)))
    ax.add_patch(ellipse)
    ax.set_axis_off()
    plt.tight_layout()
    plt.show()
    return fig, data

def sweep_rf(spiketimes, stim_data, pre=0, post=7,binsize=0.05):
    Nbins = int((post+pre)/binsize)-1
    features = np.unique(stim_data.ori)
    features.sort()
    Ntrials = int(stim_data.shape[0]/len(features))

    psth_all=np.zeros((len(features),Nbins))
    bytrial_all=np.zeros((len(features),Ntrials,Nbins))
    var_all = np.zeros((len(features),Nbins))

    for i,feat in enumerate(features):
        stmtimes = stim_data.start_time[stim_data.ori == feat].values
        psth_,bytrial,var = par.psth_arr(spiketimes,stmtimes,pre=pre,post=post,binsize=binsize,variance=True)
        psth_all[i,:] = psth_
        bytrial_all[i,:,:] = bytrial
        var_all[i,:] = var
    
    return(psth_all,bytrial_all,var_all)

def sweep_proj(data,directions,method):
    '''
    Generates back-projection map of receptive field estimation based on sweep psth data

    DATA: directions x mean psth
        NOT by trial
        Must be numpy array
    DIRECTIONS: 1D list or array of angles of presentation
    METHOD:
        0 = Arithmetic
        1 = Geometric
    '''
    
    sz =  data.shape[1]

    pad = int(np.ceil(np.sqrt(2)*sz-sz)/2)
    pad_dat = np.zeros(int(sz+2*pad))
    pad_dat[:] = np.nan

    cj = np.asmatrix(np.tile(np.arange(-(sz)/2,(sz)/2,1),(sz,1)))
    ci=cj.H

    map_ = np.ones((sz,sz))
    map_[:,:] = method

    for i in range(len(directions)):
        t = directions[i]*np.pi/180
        pad_dat[pad:pad+sz] = data[i]
        rcj = np.asmatrix(np.round(cj*np.cos(t)-ci*np.sin(t)+np.ceil(pad+sz/2)),dtype='int16')
        dat_array = pad_dat[rcj] 

        if method != 0:
            map_ = map_*dat_array
            map_ = map_/(i+1)
        else:
            map_ = map_ + dat_array

    return map_

def mask_grating(spiketimes, stim_data, xpos, ypos, pre, post, binsize):
    '''
    This function is dependent on column titles I created for my NWB files. The column of x coordinates is named "PosX" and the column of Y coordinates is named "PosY". Start time for the stimulus is named 'start_time'

    INPUTS
    spiketimes: 1D array of spike times for a single unit
    stim_data: A pandas dataframe of experiment data requiring specific column names. See above
    xpos, ypos: list of *unique* x and y positions for stimulus presentation. 
    pre, post: time before and after stim presentation for raster/psth
    binsize: desired binsize for analysis

    OUTPUTS
    labels: list of xy coordinate combinations in order of iteration
    psth_all: mean psth for the unit at a given xy coordinate
    bytrial_all: psth for each trial at a given xy 
    var_all: variance (relative to mean psth) for each unit at a given xy

    Juan Santiago 12/07/2022
    '''
    if 'start_time' and 'PosX' and 'PosY' not in stim_data.columns:
        raise Exception('stim_data requires columns named "start_time", "PosX", and "PosY"')

    Nbins = int((pre+post)/binsize-1)

    Ntrial = int(stim_data.shape[0]/(len(ypos)*len(xpos)))

    psth_all    = np.zeros(((len(ypos)*len(xpos),Nbins)))
    var_all     = np.zeros(((len(ypos)*len(xpos),Nbins)))
    bytrial_all = np.zeros((len(ypos)*len(xpos),Ntrial,Nbins))
    labels      = []

    k=0
    for i, y in enumerate(ypos):
        for j, x in enumerate(xpos):
            # Dependent on column titles
            stim_times = stim_data.start_time[stim_data.PosY == y][stim_data.PosX == x].values
            psth_,bytrial_,var_ = par.psth_arr(spiketimes,stim_times,pre=pre,post=post,binsize = binsize, variance=True)
            psth_all[k,:] = psth_
            var_all[k,:]  = var_
            bytrial_all[k,:,:] = bytrial_
            labels.append((x,y))
            k += 1

    return labels, psth_all, bytrial_all, var_all



#general imports
import os,sys,glob
import numpy as np
import pandas as pd
import h5py
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
import matplotlib.gridspec as gridspec
plt.style.use('dark_background')
import seaborn as sns
import csv 
import time, datetime

from pynwb import NWBFile
from pynwb import NWBFile, NWBHDF5IO, TimeSeries, ProcessingModule
from pynwb.device import Device
from pynwb.epoch import TimeIntervals
from pynwb.ecephys import ElectrodeGroup
from pynwb.behavior import BehavioralEvents, BehavioralEpochs, BehavioralTimeSeries, PupilTracking, IntervalSeries
   
from dateutil.tz import tzlocal
sns.set_palette(sns.color_palette('Set2')[3:])


class Session:
    """#class data object made by dan for our task. 
        #input when making a new instance is a path (string) to the data in the Deepnote file tree
        #e.g., s = session('/data/thisisafolder/thisisthefolderthecsvfilesarein')
    """
    def __init__(self,path):
        self.path = path.lower()
        print(self.path)
        self.df_pulls = pd.read_csv(os.path.join(path,'lever_pulls.csv'))
        self.df_pushes = pd.read_csv(os.path.join(path,'lever_pushes.csv'))
        self.df_releases = pd.read_csv(os.path.join(path,'lever_releases.csv'))
        self.df_rewards = pd.read_csv(os.path.join(path,'rewards.csv'),index_col=False)
        self.df_target_images = pd.read_csv(os.path.join(path,'target_images.csv'))
        self.df_trial_ends = pd.read_csv(os.path.join(path,'trial_ends.csv'))
        self.df_trial_starts = pd.read_csv(os.path.join(path,'trial_starts.csv'))
        if os.path.exists(os.path.join(path,'orientation.csv')):
            self.df_orientations = pd.read_csv(os.path.join(path,'orientation.csv'))

        self.df_pulls['lever_direction'] =  ['pull' for i in range(self.df_pulls.shape[0])]
        self.df_pushes['lever_direction'] =  ['push' for i in range(self.df_pushes.shape[0])]
        self.df_lever = pd.concat((self.df_pushes,self.df_pulls),ignore_index=True).sort_values('Frame')

    def clean_duplicates(self, df):
        return df.drop(np.argwhere(np.diff(df.Frame.values) < 5)[0][0]+1)

    def check_number_of_trials(self):
        if self.df_trial_starts.shape[0] == self.df_lever.shape[0] == self.df_releases.shape[0] == (self.df_target_images.shape[0]+1):
            self.num_trials = self.df_trial_starts.shape[0]
            print('number of trials: '+str(self.num_trials))
            return True
        else:
            # print('number of trial starts: '+str(self.df_trial_starts.shape[0]))
            # print('number of lever moves: '+str(self.df_lever.shape[0]))
            # print('number of releases: '+str(self.df_releases.shape[0]))
            # print('number of target images: '+str((self.df_target_images.shape[0]+1)))
            return False
    
    def make_df(self, phase, offset=0):
        if phase == 1:
            if not self.check_number_of_trials():
                
                if self.df_trial_starts.shape[0] == self.df_lever.shape[0]:
                    #make sure releases match by removing releases that happen within 3 frames of each other
                    self.df_releases = self.df_releases.iloc[np.where(np.diff(self.df_releases.Frame.values)>3)[0]]
                else: 
                    print(path+' trial starts and lever movements do not match, which is a problem for Phase1')
                    return

            #enforce target image shape
            self.df_target_images=self.df_target_images.iloc[np.where(self.df_target_images.Frame.values < np.max(self.df_trial_starts.Frame.values))[0]]
            
            #combine times for each trial
            self.df = pd.concat([self.df_trial_starts.drop('Value',axis=1).reset_index(),
                                self.df_releases.drop('Value',axis=1).reset_index(),
                            self.df_lever.drop('Value',axis=1).reset_index()],
                            axis=1,ignore_index=False)

            #rename columns
            self.df.columns = ['trial_start_index','trial_start_frame','trial_start_timestamp',
                        'release_index','release_frame', 'release_timestamp',
                        'lever_move_index','lever_move_frame','lever_move_timestamp','lever_move_direction']

            #put in target and possible reward information and times
            # reward_time = np.zeros(self.df.shape[0]);reward_time[:]=np.nan
            # for reward_frame in self.df_rewards.Frame.values:
            #     for trial in range(self.df.shape[0]):
            #         if abs(reward_frame - self.df['release_frame'][trial]) < 4:
            #             reward_time[trial] = self.df_rewards.Timestamp[self.df_rewards.Frame==reward_frame]
            # self.df['reward_timestamp']=reward_time
            self.df['target_image_int'] = [1]+np.array(self.df_target_images.Value).tolist()
            directions = ['','pull','push']
            image = ['cheetah','elephant']
            self.df['target_image'] = [image[i] for i in self.df['target_image_int']]
            self.df['target_direction'] = [directions[i] for i in self.df['target_image_int']]
        
            reward_,reward_times,reward_frames=[],[],[]
            for index, row in self.df.iterrows():
                start_frame =  self.df.trial_start_frame[index]
                try: end_frame = self.df.trial_start_frame[index+1]
                except:end_frame = self.df.trial_start_frame[index]+1000
                _ = gather_events_by_trial(self.df_rewards,start_frame,end_frame)
                reward_.extend([_[0]])
                reward_frames.extend([_[1]])
                reward_times.extend([_[2]])
            self.df['reward'] = reward_
            self.df['reward_times'] = reward_times
            self.df['reward_frames'] = reward_frames

        if phase == 2:
            self.df = self.df_trial_starts.drop('Value',axis=1)
            self.df.columns=['trial_start_time','trial_start_Timestamp']
            first_trial = self.df.trial_start_Timestamp.iloc[0]
            ind = np.where(first_trial > self.df_trial_ends.Timestamp.values)
            trials_no = self.df.shape[0]

            
            if self.df.trial_start_time.values[0] > self.df_trial_ends.Timestamp.values[0]:
            #     #find where the first df_trial_ends row is greater than the first trial
                # print('first start: '+str(first_trial))
                self.df_trial_ends = self.df_trial_ends[self.df_trial_ends.Timestamp > first_trial]
            
            
            if trials_no > self.df_trial_ends.shape[0]:
                print('more starts than ends')
                self.end_frames = self.df_trial_ends.Frame.values[trials_no*-1:].tolist()
                self.end_times = self.df_trial_ends.Timestamp.values[trials_no*-1:].tolist()
                self.end_Timestamps = self.df_trial_ends.Timestamp.values[trials_no*-1:].tolist() #self.df_trial_ends.Value.values[trials_no*-1:].tolist()
                for j in range(trials_no - self.df_trial_ends.shape[0]):
                    self.end_frames.extend([np.nan])
                    self.end_times.extend([np.nan])
                    self.end_Timestamps.extend([np.nan])
                self.df['trial_end_frame'] = self.end_frames
                # self.df['trial_end_time'] = self.end_times
                self.df['trial_end_Timestamp'] =  self.end_Timestamps
                self.df['trial_end_time'] = self.end_Timestamps
            else:
                self.df['trial_end_frame'] = self.df_trial_ends.Frame.values[trials_no*-1:]
                # self.df['trial_end_time'] = self.df_trial_ends.Value.values[trials_no*-1:]
                self.df['trial_end_Timestamp'] = self.df_trial_ends.Timestamp.values[trials_no*-1:]
                self.df['trial_end_time'] = self.df_trial_ends.Timestamp.values[trials_no*-1:]
            image = ['cheetah','elephant']
            
            self.df['target_image'] = [image[i] for i in self.df_target_images.Value.values]
            
            self.df['trial_start_time'] = self.df['trial_start_Timestamp']

            pull_,push_,reward_ = [],[],[]
            pull_times,push_times,reward_times= [],[],[]
            push_frames,pull_frames,reward_frames= [],[],[]
            for index, row in self.df.iterrows():
                start_frame = self.df_trial_starts.Frame.iloc[index]
                # end_frame = self.df.trial_end_frame[index+offset]
                try:
                    # end_frame = 
                    end_frame = self.df_trial_ends[self.df_trial_ends.Frame > start_frame].Frame.values[0]+2
                    #gather pulls
                    _ = gather_events_by_trial(self.df_pulls,start_frame,end_frame)
                    # print(start_frame)
                    # print(end_frame)
                    # print(_)
                    pull_.extend([_[0]])
                    pull_frames.extend([_[1]])
                    pull_times.extend([_[2]])
                    _ = gather_events_by_trial(self.df_pushes,start_frame,end_frame)
                    push_.extend([_[0]])
                    push_frames.extend([_[1]])
                    push_times.extend([_[2]])
                    _ = gather_events_by_trial(self.df_rewards,start_frame,end_frame)
                    reward_.extend([_[0]])
                    reward_frames.extend([_[1]])
                    reward_times.extend([_[2]])
                except:
                    _=[np.nan,np.nan,np.nan]
                    pull_.extend([_[0]])
                    pull_frames.extend([_[1]])
                    pull_times.extend([_[2]])
                    push_.extend([_[0]])
                    push_frames.extend([_[1]])
                    push_times.extend([_[2]])
                    _=[False,[],[]]
                    reward_.extend([_[0]])
                    reward_frames.extend([_[1]])
                    reward_times.extend([_[2]])
            self.df['pulled'] = pull_
            self.df['pushed'] = push_
            self.df['rewarded'] = reward_
            self.df['pull_times'] = pull_times
            self.df['push_times'] = push_times
            self.df['reward_times'] = reward_times
            self.df['pull_frames'] = pull_frames
            self.df['push_frames'] = push_frames
            self.df['reward_frames'] = reward_frames
            lever_move_time = []
            for trial in self.df.iterrows():
                if np.isnan(trial[1].pulled) and np.isnan(trial[1].pushed):
                    lever_move_time.extend([np.nan])
                else:
                    if trial[1].pulled:
                        if type(trial[1].pull_times) == list:
                            lever_move_time.extend([trial[1].pull_times[0]])
                        else: lever_move_time.extend([trial[1].pull_times[0]])
                    else:
                        lever_move_time.extend([trial[1].push_times[0]])
            self.df['lever_move_time']=lever_move_time

            self.df['reaction_time'] = self.df.lever_move_time - self.df.trial_start_time


        if phase > 0:
            self.mouse = self.path.split('/')[-2].lower()
            self.date = self.path.split('/')[-1]
            # dt_string= self.date.strip(self.mouse)[1:]
            dt_string = self.date[4:]
            format = "%Y_%m_%d-%H_%M_%S"
            dt_object = datetime.datetime.strptime(dt_string, format)
            self.df['phase'] = [phase]*self.df.shape[0]
            self.df['datetime'] = [dt_object]*self.df.shape[0]
            self.df['year'] = [dt_object.year]*self.df.shape[0]
            self.df['month'] = [dt_object.month]*self.df.shape[0]
            self.df['day'] = [dt_object.day]*self.df.shape[0]
            start_in_seconds = dt_object.hour*60*60 + dt_object.minute*60 + dt_object.second 
            self.df['start_in_seconds'] = [start_in_seconds]*self.df.shape[0]
            try:
                for col in self.df.columns:
                    if 'time' in col:
                        if 'trial' in col:
                            self.df[col] =  np.array(self.df[col].values) + start_in_seconds
                        else:
                            self.df[col] = [np.array(np.array(ts)+start_in_seconds).tolist() for ts in self.df[col].values]
                        # self.df[col].values+start_in_seconds
            except: pass
            try: self.df['target_image'] = [image[i] for i in self.df['target_image_int']]
            except: pass
            self.df['mouse']=[self.mouse]*self.df.shape[0]

            if os.path.exists(os.path.join(path,'orientation.csv')):
                self.df['orientations'] = [int(v.strip(')')) for v in s.df_orientations.Value.values]
        
def gather_events_by_trial(df,start_frame,end_frame):
#this function takes a DataFrame as input, and searches this DataFrame 
#for any events in the Frame column that falls between the input start_frame and end_frame 
    moved, frames, times = False, [], []
    for index, row_ in df.iterrows():
        if row_.Frame > start_frame and row_.Frame < end_frame:
            moved = True
            frames.extend([row_.Frame])
            times.extend([row_.Timestamp])
    return moved, frames, times
    
def gather_events_by_time(df,start_time,end_time):
#this function takes a DataFrame as input, and searches this DataFrame 
#for any events in the Frame column that falls between the input start_frame and end_frame 
    moved, frames, times, values = False, [], [], []
    for index, row_ in df.iterrows():
        if row_.Timestamp > start_time and row_.Timestamp < end_time:
            moved = True
            frames.extend([row_.Frame])
            times.extend([row_.Timestamp])
            values.extend([row_.Value])
    return moved, times, values

def check_session_phase(s):
    if s.df_trial_starts.shape[0] > 0: #make sure there are some trials 
        if s.df_trial_starts.shape[0] == s.df_lever.shape[0]:# in phase 1, trials are started by lever moves, so this is equal
            #could also be 2c, which is 2AFC! need to check
            return 1
        else:
            #TODO: differentiate between 2a and 2b and 3
            return 2
    else: return 0

def combine_Sessions(sessions):
#input: sessions is a list of Session objects created by dlabbehavior cheetah_elephant Session
#outpu
    sessions[0].make_df(2)
    df = sessions[0].df
    for s in sessions[1:]:
        s.make_df(2)
        df_toadd = s.df
        end_of_last_session_t = df.trial_end_time.values[-1]
        end_of_last_session_f = df.trial_end_frame.values[-1]
        for col in ['trial_start_time','trial_end_time','lever_move_time']:
            df_toadd[col] = df_toadd[col] + end_of_last_session_t
        
        for col in ['pull_times','push_times','reward_times']:
            df_toadd[col] = [np.array(np.array(times) + end_of_last_session_t).tolist for times in df_toadd[col]]
        
        df_toadd.trial_end_frame = df_toadd.trial_end_frame.values + end_of_last_session_f
        
        for col in ['pull_frames','push_frames','reward_frames']:
            df_toadd[col] = [np.array(np.array(frames) + end_of_last_session_f).tolist for frames in df_toadd[col]]
        
        df = pd.concat([df,df_toadd],ignore_index=True)
    return df

def sort_day_sessions(paths):
    new_names = []
    for path in paths:
        session_time_string = os.path.basename(path).split('-')[-1]
        if len(session_time_string.split('_')[0]) < 2:
            hour = '0'+ session_time_string.split('_')[0]
        else: hour = session_time_string.split('_')[0]
        if len(session_time_string.split('_')[1]) < 2:
            minute = '0'+ session_time_string.split('_')[1]
        else: minute = session_time_string.split('_')[1]
        if len(session_time_string.split('_')[2]) < 2:
            second = '0'+ session_time_string.split('_')[2]
        else: second = session_time_string.split('_')[2]
#os.path.basename(path).split('-')[0]+'-'+
        new_names.extend([hour+'_'+minute+'_'+second])
    return np.array(paths)[np.argsort(new_names).astype(int)]

def import_timestamps(filename):
    #returns all timestamp columns as numpy arrays
    #input: path to csv with timestamps in a column called 'Timestamp'
    df1 = pd.read_csv(filename) 
    df = pd.to_numeric(df1['Timestamp'])                                  #These are actually timestamps
    timestamps = np.array(df)                                             #should be in seconds
    return timestamps

def make_df(path,rew_amt=-1):
    #Reads all timestamps from each csv from a session
    # filename = os.path.join(path,"trial_starts.csv")                  #should be when images came on
    # df_trial_starts = pd.read_csv(filename,index_col=False) 
    # trial_starts = np.array(df_trial_starts.Timestamp)
    
    filename = os.path.join(path,"lever_pulls.csv")
    df_lever_pulls = pd.read_csv(filename,index_col=False) 
    lever_pulls =  np.array(df_lever_pulls.Timestamp)
    
    filename = os.path.join(path, "lever_releases.csv")
    df_lever_releases = pd.read_csv(filename,index_col=False) 
    lever_releases =  np.array(df_lever_releases.Timestamp)

    filename = os.path.join(path, "trial_ends.csv")
    df_trial_ends = pd.read_csv(filename,index_col=False) 
    trial_ends =  np.array(df_trial_ends.Timestamp) 

    filename = os.path.join(path, "manual_rewards.csv")
    df_manual_rewards = pd.read_csv(filename,index_col=False) 
    manual_rewards = np.array(df_manual_rewards.Timestamp) 

    filename = os.path.join(path, "rewards.csv")
    df_rewards = pd.read_csv(filename,index_col=False) 
    df_rewards_withAmount = pd.read_csv(filename) 
    df_rewards['Value']=[int(amt.strip(')')) for amt in df_rewards_withAmount.Value]
    rewards = np.array(df_manual_rewards.Timestamp)

    filename = os.path.join(path, "target_times.csv")
    if os.path.exists(filename):
        df_target_times = pd.read_csv(filename,index_col=False) 

    ###Not Global Variables
    new_rew_amt = rew_amt

    ##################Makes df
    session_start = lever_pulls[0] #timestamp of first pull

    rewardsZero = rewards - session_start - .002

    lever_pullsZero = lever_pulls - session_start
    lever_releasesZero = lever_releases - session_start
    trial_endsZero = trial_ends - session_start
    # trial_startsZero = trial_starts - session_start

    filename = os.path.join(path, "target_times.csv")
    if os.path.exists(filename):
        target_timestamp = df_target_times.Timestamp - session_start
        a=df_target_times.Value[:-1].to_list()
        a.insert(0,1.0)
        target_times = a
    else:
        target_timestamp = [np.nan]*np.shape(lever_pulls)[0]
        target_times = target_timestamp
 
    if len(manual_rewards):
        manual_rewardsZero = manual_rewards - session_start 

    #Checking for alignment
    new_lever_pullsZero = lever_pullsZero
    new_lever_pullsZero1 = new_lever_pullsZero
    if len(lever_releasesZero) == len(lever_pullsZero) and np.min(lever_releasesZero-lever_pullsZero)>0:
        # plt.plot(lever_releasesZero-lever_pullsZero)
        #no alignment necessary
        new_lever_releasesZero = lever_releasesZero
        new_lever_releasesZero1 = lever_releasesZero
    else: 
        #aligment here, removes last lever pull if the session ends before release happens 
        new_lever_releasesZero = []
        for pull in lever_pullsZero:
            pullfound=False
            for release in lever_releasesZero: 
                if not pullfound:
                    if release > pull:
                        new_lever_releasesZero.extend([release])
                        pullfound=True

        if np.shape(new_lever_releasesZero)[0] + 1 == np.shape(new_lever_pullsZero)[0]:
            new_lever_releasesZero.extend([np.nan])
        new_lever_releasesZero1 = new_lever_releasesZero

    #Use these for properly aligned pulls/releases per trial:
    new_lever_pullsZeroArray = np.array(new_lever_pullsZero)  
    new_lever_releasesZeroArray = new_lever_releasesZero1

    #calculate the hold time
    actual_holdTime = new_lever_releasesZeroArray -  new_lever_pullsZeroArray

    #get the rewarded trials
    rewarded = [gather_events_by_time(df_rewards,start+session_start,end+session_start+0.200)[0] for start,end in zip(new_lever_pullsZeroArray,new_lever_releasesZeroArray)]

    # get the reward amounts
    # Keep in mind that only the rewarded lines will have the vol of reward
    reward_amount = []
    for start,end in zip(new_lever_pullsZeroArray,new_lever_releasesZeroArray):
        val = gather_events_by_time(df_rewards,start+session_start,end+session_start+0.200)[2]
        if len(val)>0: val = sum(val) #there was a reward, note the value 
        else: val = 0. #there was no reward, note the reward amount as 0
        reward_amount.extend([val])


    #for phase 1 and 2, create dummy targets
    if np.shape(target_timestamp)[0]==np.shape(rewarded)[0]:pass
    else:
        target_timestamp = [np.nan]*np.shape(rewarded)[0]
        target_times = target_timestamp

    #%% Construct the df:
    all_data = list(zip(new_lever_pullsZero, new_lever_releasesZero, actual_holdTime, rewarded, reward_amount,target_timestamp,target_times)) #, dip, p_dip))
    all_data_df = pd.DataFrame(all_data, columns=['pull_times', 'release_times', 'hold_time', 'rewarded','reward_amount','target_timestamp','target_hold_time'])#, 'dip cutoff', 'p_dip'])
    all_data_df['reaction_time']=all_data_df.hold_time-all_data_df.target_hold_time

    return all_data_df.iloc[1:]#throw out the first trial

def make_lever_NWB(recording_folder,experimenter='djd',experiment_description= 'Denman Lab lever visual engagement task',reward_vol=-1):
    nwb_path = os.path.join(recording_folder,os.path.basename(recording_folder))+'_lever.nwb'
    
    if os.path.exists(os.path.join(recording_folder,'image_ends.csv')):
        df_target = pd.read_csv(os.path.join(recording_folder,'target_times.csv'))
        if df_target.shape[0]<1 : phase = 2
        else: phase = 3
    else:
        phase=1

    if not os.path.exists(nwb_path):
        try:
            df = make_df(recording_folder)
        except:
            print('FAILED: '+recording_folder+' at df creation')
            return np.nan
        try:
            nwb_file = NWBFile(experiment_description, 
                    recording_folder, 
                    datetime.datetime.now(tzlocal()),
                    experimenter=experimenter,
                    lab='Denman Lab',
                    institution='University of Colorado',
                    experiment_description=experiment_description+' | phase:'+str(phase),
                    session_id=os.path.basename(recording_folder))
        except:
            print('FAILED: '+recording_folder+' at file creation')
            return np.nan
        try:
            behavior_module = ProcessingModule('behavior', 'behavior module')
            nwb_file.add_processing_module(behavior_module)
        except:
            print('FAILED: '+recording_folder+' at behavior module creation')
            return np.nan
        try:
            nwb_file.add_trial_column('hold_time','time in seconds the mouse held the lever for')
            nwb_file.add_trial_column('target_time','time in seconds that the mouse needed to hold the lever for.')
            nwb_file.add_trial_column('rewarded','if a reward was delivered on this trial')
            nwb_file.add_trial_column('reward_amount','amount in uL of the reward')
            
            if not 'target_time' in df.columns:
                df['target_time']=[np.nan]*df.shape[0]
            for i,unit_row in df.iterrows():
                nwb_file.add_trial(start_time=unit_row.pull_times,
                                stop_time=float(unit_row.release_times),
                                hold_time=unit_row.hold_time,
                                target_time=unit_row.target_time,
                                reward_amount=unit_row.reward_amount,
                                rewarded=unit_row.rewarded)
        except:
            print('FAILED: '+recording_folder+' at adding trials')
            return np.nan
        try:
            l = pd.read_csv(os.path.join(recording_folder,'lever.csv'),skiprows=1,header=None,index_col=False)
            start = pd.read_csv(os.path.join(recording_folder,'session_start.csv'),header=None,index_col=False)
            start_time = start.iloc[0].values[0].split('T')[-1].split('-')[0]
            l['counter'] = np.array([int(count[-13:-2]) if type(count)==str else np.nan for count in l[6]])/1e7
            l['time'] = [str(int(l[3].iloc[i]))+':'+str(int(l[4].iloc[i]))+':'+str(int(l[5].iloc[i]))+'.'+str(l[2].iloc[i])[2:] if not np.isnan(l[5].iloc[i]) else np.nan for i in range(l.shape[0])]
            lever_start = float(l.time.iloc[0].split(':')[-1]) - float(start_time.split(':')[-1])
            lever_offset = l['counter'].iloc[0] - lever_start
            l['Timestamp'] = l['counter'].values - lever_offset
            
            val_ = [int(vala[1:]) if  type(vala)==str else np.nan for vala in l[0]]
            l['Value'] = val_

            lever_ts = TimeSeries(
                name='lever_position',
                timestamps=l['counter'].values - lever_offset,
                data=val_,
                unit='arbitrary',
                description='The position reading of the 4 way encoder attached to the lever the mouse pushes forward, backward, left and right with his forelimbs.',
                comments='Lever movement. 512 is centered.'
            )
            nwb_file.add_acquisition(lever_ts)
        except: print('FAILED: '+recording_folder+' at lever data')
        try:
            with NWBHDF5IO(nwb_path, 'w') as io:
                io.write(nwb_file)
            print('SUCCEEDED: '+recording_folder)

        except:print('FAILED: '+recording_folder+' at writing')
    else: print('EXISTS: '+os.path.join('/root/work/nwbs_lever',os.path.basename(recording_folder))+'_lever.nwb')

def make_lever_NWB_verbose(recording_folder,experimenter='djd',experiment_description= 'Denman Lab lever visual engagement task',reward_vol=-1):
    if not os.path.exists(os.path.join(recording_folder,os.path.basename(recording_folder))+'_lever.nwb'):

        df = make_df(recording_folder,reward_vol)

        nwb_file = NWBFile(experiment_description, 
                recording_folder, 
                datetime.datetime.now(tzlocal()),
                experimenter=experimenter,
                lab='Denman Lab',
                institution='University of Colorado',
                experiment_description=experiment_description,
                session_id=os.path.basename(recording_folder))

        behavior_module = ProcessingModule('behavior', 'behavior module')
        nwb_file.add_processing_module(behavior_module)

        nwb_file.add_trial_column('hold_time','time in seconds the mouse held the lever for')
        nwb_file.add_trial_column('target_time','time in seconds that the mouse needed to hold the lever for.')
        nwb_file.add_trial_column('rewarded','if a reward was delivered on this trial')
        nwb_file.add_trial_column('reward_amount','amount in uL of the reward')
        
        if not 'target_time' in df.columns:
            df['target_time']=[np.nan]*df.shape[0]
        for i,unit_row in df.iterrows():
            nwb_file.add_trial(start_time=unit_row.pull_times,
                            stop_time=unit_row.release_times,
                            hold_time=unit_row.hold_time,
                            target_time=unit_row.target_time,
                            reward_amount=unit_row.hold_time,
                            rewarded=unit_row.rewarded)

        l = pd.read_csv(os.path.join(recording_folder,'lever.csv'),skiprows=1,header=None,index_col=False)
        start = pd.read_csv(os.path.join(recording_folder,'session_start.csv'),header=None,index_col=False)
        start_time = start.iloc[0].values[0].split('T')[-1].split('-')[0]
        l['counter'] = np.array([int(count[-13:-2]) for count in l[6]])/1e7
        l['time'] = [str(l[3].iloc[i])+':'+str(l[4].iloc[i])+':'+str(l[5].iloc[i])+'.'+str(l[2].iloc[i])[2:] for i in range(l.shape[0])]
        lever_start = float(l.time.iloc[0].split(':')[-1]) - float(start_time.split(':')[-1])
        lever_offset = l['counter'].iloc[0] - lever_start
        l['Timestamp'] = l['counter'].values - lever_offset
        
        val_ = [int(vala[1:])for vala in l[0]]
        l['Value'] = val_
        
        lever_ts = TimeSeries(
            name='lever_position',
            timestamps=l['counter'].values - lever_offset,
            data=val_,
            unit='arbitrary',
            description='The position reading of the 4 way encoder attached to the lever the mouse pushes forward, backward, left and right with his forelimbs.',
            comments='Lever movement. 512 is centered.'
        )
        nwb_file.add_acquisition(lever_ts)

        nwb_path = os.path.join('nwbs_lever',os.path.basename(recording_folder))+'_lever.nwb'
        with NWBHDF5IO(nwb_path, 'w') as io:
            io.write(nwb_file)
        print('SUCCEEDED: '+recording_folder)

def make_discrimination_NWB(recording_folder,experimenter='djd',experiment_description= 'Denman Lab visual engagement task',reward_vol=-1):
    nwb_path = os.path.join('/root/work/nwbs',os.path.basename(recording_folder))+'_ce.nwb'
    if not os.path.exists(nwb_path):
        print('making discimination NWB for '+recording_folder)
        try:
            s = Session(recording_folder)
            s.make_df(2)
            if s.df.trial_start_Timestamp.iloc[0] > s.df.trial_end_Timestamp.iloc[0]:
                print('FAILED: '+recording_folder+' at df alignment')
                print(s.df.trial_start_time.iloc[0]);print(s.df.trial_end_time.iloc[0])

        except:
            print('FAILED: '+recording_folder+' at df creation')
        try:
            # y,m,d = folders[0].split('/')[-1].split('-')[0].split('_')[1:] 
            # if int(y) == 2021 and int(m) > 9 and int(d) > 13: experiment_description= 'Denman Lab plaid discrimination'
            # else: experiment_description = 'Denman Lab cheetah or elephant discrimination'
            nwb_file = NWBFile(experiment_description, 
                    recording_folder, 
                    datetime.datetime.now(tzlocal()),
                    experimenter=experimenter,
                    lab='Denman Lab',
                    institution='University of Colorado',
                    experiment_description=experiment_description,
                    session_id=os.path.basename(recording_folder))
        except:print('FAILED: '+recording_folder+' at file creation')
        try:
            behavior_module = ProcessingModule('behavior', 'behavior module')
            nwb_file.add_processing_module(behavior_module)
        except:print('FAILED: '+recording_folder+' at behavior module creation')
        # try:s
        if 'cheetah' in experiment_description: 
            nwb_file.add_trial_column('image','image presented on this trial.')
            nwb_file.add_trial_column('percent_elephant','contrast of one grating presented on this trial.')
            nwb_file.add_trial_column('pulled','time in seconds that the trial ended.')
            nwb_file.add_trial_column('pushed','time in seconds that the trial ended.')
            nwb_file.add_trial_column('response_time','time in seconds that the trial ended.')
            nwb_file.add_trial_column('rewarded','if a reward was delivered on this trial')

            for i,unit_row in s.df.iterrows():
                if unit_row.target_image == 'elephant':
                    percent_elephant= 100; 
                else: 
                    percent_elephant= 0; 
                nwb_file.add_trial(start_time=unit_row.trial_start_Timestamp,
                                stop_time=float(unit_row.trial_end_Timestamp),
                                image=unit_row.target_image,
                                percent_elephant=str(percent_elephant),
                                pulled=str(unit_row.pulled),
                                pushed=str(unit_row.pushed),
                                response_time=str(unit_row.lever_move_time),
                                rewarded=unit_row.rewarded)
        if 'plaid'in experiment_description:
            nwb_file.add_trial_column('image','image presented on this trial.')
            nwb_file.add_trial_column('imageApercent','contrast of one grating presented on this trial.')
            nwb_file.add_trial_column('imageBpercent','contrast of one grating presented on this trial.')
            nwb_file.add_trial_column('pulled','time in seconds that the trial ended.')
            nwb_file.add_trial_column('pushed','time in seconds that the trial ended.')
            nwb_file.add_trial_column('response_time','time in seconds that the trial ended.')
            nwb_file.add_trial_column('rewarded','if a reward was delivered on this trial')
            # if 'orientation' in s.df.columns:
            nwb_file.add_trial_column('orientation','orientation of grating')


            for i,unit_row in s.df.iterrows():
                if unit_row.target_image == 'elephant':
                    percA= 0; percB= 100;
                else: 
                    percA= 100; percB= 0;
                if 'orientation' in s.df.columns:
                    ori = unit_row.orientation
                else: ori=np.nan
                nwb_file.add_trial(start_time=float(unit_row.trial_start_Timestamp)+s.df.start_in_seconds.iloc[0],
                                stop_time=float(unit_row.trial_end_Timestamp)+s.df.start_in_seconds.iloc[0],
                                image='grating',
                                imageApercent=str(percA),
                                imageBpercent=str(percB),
                                pulled=str(unit_row.pulled),
                                pushed=str(unit_row.pushed),
                                response_time=str(unit_row.lever_move_time),
                                rewarded=unit_row.rewarded,
                                orientation=ori)



        # except:print('FAILED: '+recording_folder+' at adding trials')
        # try:
        lf = pd.read_csv(os.path.join(recording_folder,'buffered_lever_frames.csv'),index_col=False)

        val_ = [int(vala[1:]) if  type(vala)==str else np.nan for vala in lf.Value]
        lf.Value = val_
        
        lever_ts = TimeSeries(
            name='lever_position',
        #     starting_time=lf.Timestamp.iloc[0],
            timestamps=lf.Timestamp.values+s.df.start_in_seconds.iloc[0],
            data=lf.Value.values,
            unit='arbitrary',
            description='The position reading of the 4 way encoder attached to the lever the mouse pushes forward, backward, left and right with his forelimbs.',
            comments='Lever movement. 512 is centered.'
        )
        nwb_file.add_acquisition(lever_ts)
        # except: 
        #     print('FAILED: '+recording_folder+' at lever data')
        if os.path.exists(os.path.join(recording_folder,'spike_clusters.npy')): #if there is sorted ephys data
            nwb_file = add_ephys_data(nwb_file,recording_folder)
        
        try:
            nwb_path = os.path.join('/root/work/nwbs',os.path.basename(recording_folder))+'_ce.nwb'
            with NWBHDF5IO(nwb_path, 'w') as io:
                io.write(nwb_file)
            print('SUCCEEDED: '+recording_folder)
        except:print('FAILED: '+recording_folder+' at writing')

def make_behavior_NWB(recording_folder,experimenter,experiment_description,reward_vol=-1):
    if 'plaid' or 'cheetah' in experiment_description:
        make_discrimination_NWB(recording_folder,experimenter,experiment_description,reward_vol)
    if 'lever' in experiment_description:
        make_lever_NWB(recording_folder,experimenter,experiment_description)

def combine_nwb_sessions(paths):
    io = NWBHDF5IO(paths[0], mode='r')
    nwb_ = io.read()
    df = nwb_.trials.to_dataframe()
    for nwb_path in paths[1:]:
        io = NWBHDF5IO(nwb_path, mode='r')
        nwb_ = io.read()
        df_toadd = nwb_.trials.to_dataframe()
        end_of_last_session_t = df.stop_time.values[-1]+10. #offset 10 seconds by default. this is inaccurate but a minimum estimate
        for col in ['start_time','stop_time','response_time']:
            if col in df_toadd.columns:
                df_toadd[col] = df_toadd[col].values.astype(float) + end_of_last_session_t        
        df = pd.concat([df,df_toadd],ignore_index=True)
    return df

def across_session_plots_lever(df):
    fig = plt.figure(figsize=(8.5,11))
    ax_num_trials = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.05,.03),yspan=(0.1,0.34))
    ax_rewards = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.37,0.62),yspan=(0.1,0.34))
    ax_hold_times_hist = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.05,0.95),yspan=(0.4,0.95))
    ax_percent_correct = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.67,0.92),yspan=(0.1,0.34))

    num_trials,rewards, reward_vol,percent_correct = [],[],[],[]
    for session_ in df.session_id.unique():
        df_sess = df[df.session_id==session_]
        num_trials.extend([df_sess.shape[0]])
        rewards.extend([sum(df_sess.rewarded)])
        reward_vol.extend([sum(df_sess.reward_amount)])
        percent_correct.extend([sum(df_sess.rewarded) / float(df_sess.shape[0])])

        ax=sns.histplot(data=df_sess, x="hold_time",hue='rewarded',binrange=(0,12),bins=60,
                    ax=ax_hold_times_hist)

    ax_num_trials.plot(df.session_id.unique(),num_trials)  
    ax_rewards.plot(df.session_id.unique(),rewards,label='rewards',color=sns.color_palette[3])  
    ax_rewards.twinx().plot(df.session_id.unique(),reward_vol,label='reward volume',color=sns.color_palette[3])   
    ax_rewards.legend(bbox_to_anchor= (0.2, -0.2) )
    ax_percent_correct.plot(df.session_id.unique(),percent_correct)  
    
    return fig,ax

def generate_session_lever(mouse,date,return_=None,session='combine', nwb=None):
    if nwb!=None:
        paths=[nwb]
        io = NWBHDF5IO(nwb, mode='r')
        nwb_ = io.read()
        df = nwb_.trials.to_dataframe()
    else:
        nwb_folder_paths = glob.glob(os.path.join(r'C:\Users\denma\Desktop\bonsai_levertask\data',mouse,mouse+'_'+date+'*'))
        paths = [glob.glob(os.path.join(p,'*.nwb'))[0] for p in nwb_folder_paths]
        if len(paths)> 1:
            if session == 'combine':
                paths = sort_day_sessions(paths)
                df = combine_nwb_sessions(paths)
                io = NWBHDF5IO(paths[-1], mode='r')
                nwb_ = io.read()
            else:
                nwb_path = paths[session]
                io = NWBHDF5IO(nwb_path, mode='r')
                nwb_ = io.read()
                df = nwb_.trials.to_dataframe()
        else: 
            nwb_path = paths[0]
            io = NWBHDF5IO(nwb_path, mode='r')
            nwb_ = io.read()
            df = nwb_.trials.to_dataframe()

    if return_ == 'df':
        df['mouse_id']   = [mouse]*df.shape[0]
        df['session_id'] = [date]*df.shape[0]
        df['phase'] = [int(nwb_.experiment_description.split('phase')[1][1:])]*df.shape[0]
        return df

    if return_ == 'fig':
        fig = plt.figure(figsize=(8.5,11))
        ax_lever_all = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.0,1.0),yspan=(0,0.17))
        ax_lever_ex1 = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.0,0.2),yspan=(0.24,0.34))
        ax_lever_ex2 = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.26,0.46),yspan=(0.24,0.34))
        ax_lever_start = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.52,0.72),yspan=(0.24,0.34))
        ax_lever_end = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.78,1.0),yspan=(0.24,0.34))        
        ax_hold_times_hist = placeAxesOnGrid(fig,dim=(1,1),xspan=(0,0.25),yspan=(0.4,0.62))
        ax_hold_times_box1 = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.35,0.6),yspan=(0.4,0.62))
        ax_hold_times_box2 = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.7,0.95),yspan=(0.4,0.62))
        ax_hold_times_session = placeAxesOnGrid(fig,dim=(1,1),xspan=(0,0.42),yspan=(0.7,0.85))
        ax_hold_times_rolling = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.56,0.99),yspan=(0.7,0.85))
        ax_hold_times_rewards = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.56,0.99),yspan=(0.9,0.99))
        
        #plot lever - whole session
        if len(paths)> 1:
            if session == 'combine':
                paths = sort_day_sessions(paths)
                io = NWBHDF5IO(paths[0], mode='r')
                nwb_ = io.read()
                ax_lever_all.plot(nwb_.acquisition['lever_position'].timestamps,
                        nwb_.acquisition['lever_position'].data,'w')
                ax_lever_all.set_ylabel('lever position')
                ax_lever_all.set_xlabel('sesssion time (sec)')    
                end_time = nwb_.acquisition['lever_position'].timestamps[-1]
                for path in paths[1:]:
                    io = NWBHDF5IO(path, mode='r')
                    nwb_ = io.read()
                    ax_lever_all.plot(nwb_.acquisition['lever_position'].timestamps+end_time,
                            nwb_.acquisition['lever_position'].data,'w')
                    ax_lever_all.set_ylabel('lever position')
                    ax_lever_all.set_xlabel('sesssion time (sec)')   
                    end_time += nwb_.acquisition['lever_position'].timestamps[-1]
        else:
            ax_lever_all.plot(nwb_.acquisition['lever_position'].timestamps,
                    nwb_.acquisition['lever_position'].data,'w')
            ax_lever_all.set_ylabel('lever position')
            ax_lever_all.set_xlabel('sesssion time (sec)')
        
        #plot some zoomed in lever examples
        r = np.random.randint(np.array(nwb_.acquisition['lever_position'].timestamps).shape[0])
        ax_lever_ex2.plot(nwb_.acquisition['lever_position'].timestamps[r:r+180],
                    nwb_.acquisition['lever_position'].data[r:r+180],color='w')
        r = np.random.randint(np.array(nwb_.acquisition['lever_position'].timestamps).shape[0])
        ax_lever_ex1.plot(nwb_.acquisition['lever_position'].timestamps[r:r+180],
                    nwb_.acquisition['lever_position'].data[r:r+180],color='w')

        #plot lever around pulls and rewards
        lever_at_pull   = []
        lever_at_release = []
        for i,row in df.iterrows():
            start_indices = np.where((np.array(nwb_.acquisition['lever_position'].timestamps) > row.start_time - 0.5) & ((np.array(nwb_.acquisition['lever_position'].timestamps) < row.start_time + 1.5)))
            # start_indices = np.arange(np.where(np.array(nwb_.acquisition['lever_position'].timestamps) > row.start_time)[0][0] - 30, np.where(np.array(nwb_.acquisition['lever_position'].timestamps) > row.start_time)[0][0] + 90)
            if len(nwb_.acquisition['lever_position'].data[start_indices]) > 1:
                lever_at_pull.append(nwb_.acquisition['lever_position'].data[start_indices])
                ax_lever_start.plot(np.array(nwb_.acquisition['lever_position'].timestamps[start_indices])-row.start_time,nwb_.acquisition['lever_position'].data[start_indices],color='#808080',lw=0.5)
            
            end_indices = np.where((np.array(nwb_.acquisition['lever_position'].timestamps) > row.stop_time - 0.5) & ((np.array(nwb_.acquisition['lever_position'].timestamps) < row.stop_time + 1.5)))
            # end_indices = np.arange(np.where(np.array(nwb_.acquisition['lever_position'].timestamps) > row.stop_time)[0][0] - 30, np.where(np.array(nwb_.acquisition['lever_position'].timestamps) > row.stop_time)[0][0] + 90)
            if len(nwb_.acquisition['lever_position'].data[end_indices]) > 1:
                ax_lever_end.plot(np.array(nwb_.acquisition['lever_position'].timestamps[end_indices])-row.stop_time,nwb_.acquisition['lever_position'].data[end_indices],color='#808080',lw=0.5)
                lever_at_release.append(nwb_.acquisition['lever_position'].data[end_indices])
        ax_lever_start.set_title('lever at start of trials');ax_lever_start.set_ylabel('lever position');ax_lever_start.set_xlabel('time (sec)')
        ax_lever_end.set_title('lever at end of trials');ax_lever_end.set_ylabel('lever position');ax_lever_end.set_xlabel('time (sec)')
    
        
        # histogram of hold times 
        sns.histplot(data=df, x="hold_time",hue='rewarded',binrange=(0,12),bins=60,
                    ax=ax_hold_times_hist)

        #box plots of hold times
        ax = sns.boxplot(data=df,y='hold_time',x='rewarded',ax=ax_hold_times_box1)
        ax = sns.stripplot(data=df,y='hold_time',x='rewarded',ax=ax_hold_times_box1)
        # ax_hold_times_box1.set_ylim(0,12)


        ax = sns.boxplot(data=df,y='hold_time',x='rewarded',ax=ax_hold_times_box2)
        ax = sns.stripplot(data=df,y='hold_time',x='rewarded',ax=ax_hold_times_box2)
        # ax_hold_times_box2.set_ylim(0.01,12)
        ax_hold_times_box2.set_yscale('log')

        ax = sns.regplot(x="start_time", y="hold_time", data=df[df.rewarded==False],ax=ax_hold_times_session)
        ax = sns.regplot(x="start_time", y="hold_time", data=df[df.rewarded==True],ax=ax_hold_times_session)
        # plt.ylim(0,4)
        ax_hold_times_session.set_xlabel('session time (secs)')
        ax_hold_times_session.set_ylabel('hold time (secs)')

        ax_hold_times_rolling.plot(df[df.rewarded==False].start_time,
        df[df.rewarded==False].hold_time.rolling(5).mean())
        ax_hold_times_rolling.plot(df[df.rewarded==True].start_time,
        df[df.rewarded==True].hold_time.rolling(5).mean())
        ax_hold_times_rolling.plot(df.start_time,df.hold_time.rolling(5).mean())
        ax_hold_times_rolling.set_xlabel('session time (secs)')
        ax_hold_times_rolling.set_ylabel('hold time (secs)')

        minutes = np.arange(int(df.start_time.values[-1]/60))
        rewards_min = []
        for i in minutes:
            df_ = df[df.rewarded==True]
            rewards_min.extend([df_[(df_.start_time > i*60.) & (df_.start_time < i*60.+60)].shape[0]])
        seconds = minutes * 60.
        ax_hold_times_rewards.plot(seconds,rewards_min)
        ax_hold_times_rewards.set_xlabel('session time (secs)')
        ax_hold_times_rewards.set_ylabel('rewards per minute')

        fig.text(0.02,0.99, mouse, fontsize=14)
        fig.text(0.08,0.99, date, fontsize=14)
        fig.text(0.02,0.96, 'total rewards: '+str(df[df.rewarded==True].shape[0]), fontsize=14)
        fig.text(0.02,0.93, 'mean hold time: '+str(np.nanmean(df.hold_time)), fontsize=14)
        fig.text(0.02,0.90, 'mean rewarded time: '+str(np.nanmean(df[df.rewarded==True].hold_time)), fontsize=14)


        return fig,ax

def generate_session_plaid(mouse,date,return_=None,session='combine',log_hist=True):
    nwb_folder_paths = glob.glob(os.path.join(r'C:\Users\denma\Desktop\cheetah_or_elephant\data',mouse,mouse+'_'+date+'*'))
    paths = [glob.glob(os.path.join(p,'*.nwb'))[0] for p in nwb_folder_paths]
    if session == 'combine':
        paths = sort_day_sessions(paths);path=paths[0]
        io = NWBHDF5IO(path, mode='r')
        nwb_ = io.read()
        df = nwb_.trials.to_dataframe()
        for path in paths[1:]:
            df_ = NWBHDF5IO(path, mode='r').read().trials.to_dataframe()
            df = pd.concat([df,df_,],ignore_index=True)
    else:
        if len(paths) == 1:
            path = paths[0]
        if type(session)==int:
            path=paths[session]
        else: 
            print('incorrect format of session; provide either an int index to a sorted list of this days sessions, or str(combine) to combine all of this days sessions.')
            return np.nan
        df = NWBHDF5IO(path, mode='r').read().trials.to_dataframe()

    df['reaction_time']=df.response_time.values.astype(float) - df.start_time.values.astype(float)
    df['target_image'] = ['A' if row.imageApercent > row.imageBpercent else 'B' for i,row in df.iterrows()]

    df['mouse_id']   = [mouse]*df.shape[0]
    df['session_id'] = [date]*df.shape[0]
    # df['phase'] = [int(nwb_.experiment_description.split('phase')[1][1:])]*df.shape[0]
   
    if return_ == 'df':
        return df

    if return_ == 'fig':
        fig = plt.figure(figsize=(8.5,11))
        ax_lever_all = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.0,1.0),yspan=(0,0.12))
        ax_lever_ex1 = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.0,0.2),yspan=(0.18,0.3))
        ax_lever_ex2 = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.26,0.46),yspan=(0.18,0.3))
        ax_lever_image = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.57,1.0),yspan=(0.2,0.35))
        ax_lever_response = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.57,1.0),yspan=(0.44,0.59))        
        ax_correct_rolling = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.,0.43),yspan=(0.39,0.59))
        ax_rt = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.,0.43),yspan=(0.6,0.8))
        ax_rt_hist = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.52,0.67),yspan=(0.66,0.8))
        ax_rt_violin = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.78,1.0),yspan=(0.66,0.8))

        if len(paths) == 1 or type(session)==int:
            nwb_ = NWBHDF5IO(path, mode='r').read()
            lever_timestamps = np.array(nwb_.acquisition['lever_position'].timestamps)
            lever_data = np.array(nwb_.acquisition['lever_position'].data)
        else:
            if session == 'combine':
                print('combining '+mouse+' sessions from '+date+'...')
                lever_timestamps = []
                lever_data =      []
                for path in paths:
                    nwb_ = NWBHDF5IO(path, mode='r').read()
                    lever_timestamps.extend(nwb_.acquisition['lever_position'].timestamps)
                    lever_data.extend(nwb_.acquisition['lever_position'].data)
                lever_timestamps = np.array(lever_timestamps)
                lever_data = np.array(lever_data)
            else:print('incorrect format of session; provide either an int index to a sorted list of this days sessions, or str(combine) to combine all of this days sessions.')
        ax_lever_all.plot(lever_timestamps,lever_data,'w')
        ax_lever_all.set_ylabel('lever position')
        ax_lever_all.set_xlabel('sesssion time (sec)')  
            
        # #plot some zoomed in lever examples
        r = np.random.randint(lever_timestamps.shape[0])
        ax_lever_ex2.plot(lever_timestamps[r:r+180],
                    lever_data[r:r+180],color='w')
        r = np.random.randint(lever_timestamps.shape[0])
        ax_lever_ex1.plot(lever_timestamps[r:r+180],
                    lever_data[r:r+180],color='w')
        ylim = 30
        for ax in (ax_lever_ex1,ax_lever_ex2): ax.set_ylim(512-ylim,512+ylim)

        # #plot lever around images A and B
        A_levers = np.zeros((df[df.imageApercent > df.imageBpercent].shape[0],180));al=0
        B_levers = np.zeros((df[df.imageBpercent > df.imageApercent].shape[0],180));bl=0
        A_images = np.zeros((df[df.imageApercent > df.imageBpercent].shape[0],180));ai=0
        B_images = np.zeros((df[df.imageBpercent > df.imageApercent].shape[0],180));bi=0
        for i,row in df.iterrows():
            try:
                lever_trigger = np.where((lever_timestamps > float(row.response_time)))[0][0]
                image_trigger = np.where((lever_timestamps > float(row.start_time)))[0][0]
                lever_response_d = lever_data[lever_trigger-120:lever_trigger+60] 
                lever_response_t = lever_timestamps[lever_trigger-120:lever_trigger+60] 
                lever_image_d = lever_data[image_trigger-30:image_trigger+150] 
                lever_image_t = lever_timestamps[image_trigger-30:image_trigger+150] 
                if row.imageApercent > row.imageBpercent:
                    ax_lever_image.plot(np.linspace(-0.5,2.5,180),lever_image_d,lw=0.5,color=sns.color_palette()[0],alpha=0.3)
                    A_levers[al,:]=lever_image_d; al+=1
                else:
                    ax_lever_image.plot(np.linspace(-0.5,2.5,180),lever_image_d,lw=0.5,color=sns.color_palette()[1],alpha=0.3)
                    B_levers[bl,:]=lever_image_d; bl+=1     
                
                if row.imageApercent > row.imageBpercent:
                    ax_lever_response.plot(np.linspace(-2,1,180),lever_response_d,lw=0.5,color=sns.color_palette()[0],alpha=0.3)
                    A_images[ai,:]=lever_response_d; ai+=1
                else:
                    ax_lever_response.plot(np.linspace(-2,1,180),lever_response_d,lw=0.5,color=sns.color_palette()[1],alpha=0.3)
                    B_images[bi,:]=lever_response_d; bi+=1
            except:
                if row.imageApercent > row.imageBpercent:
                    A_levers = A_levers[:-1,:]
                    A_images = A_images[:-1,:]
                else:
                    B_levers = B_levers[:-1,:]
                    B_images = B_images[:-1,:]
            
        ax_lever_image.plot(np.linspace(-0.5,2.5,180),np.mean(A_levers,axis=0),lw=2,color=sns.color_palette()[0])
        ax_lever_image.plot(np.linspace(-0.5,2.5,180),np.mean(B_levers,axis=0),lw=2,color=sns.color_palette()[1])
        ax_lever_response.plot(np.linspace(-2,1,180),np.mean(A_images,axis=0),lw=2,color=sns.color_palette()[0])
        ax_lever_response.plot(np.linspace(-2,1,180),np.mean(B_images,axis=0),lw=2,color=sns.color_palette()[1])
            
        ax_lever_image.set_title('aligned to image onset')
        ax_lever_response.set_title('aligned lever movement')
        for ax in [ax_lever_image,ax_lever_response]:
            ax.set_xlabel('time (sec)',fontsize=10)
            ax.set_ylabel('lever (arb)',fontsize=10)

        window = 10
        ax_correct_rolling.plot(df[df.imageApercent > df.imageBpercent].start_time,
        df[df.imageApercent > df.imageBpercent].rewarded.rolling(window).mean(),color=sns.color_palette()[0])
        ax_correct_rolling.plot(df[df.imageBpercent > df.imageApercent].start_time,
        df[df.imageBpercent > df.imageApercent].rewarded.rolling(window).mean(),color=sns.color_palette()[1])
        ax_correct_rolling.plot(df.start_time,df.rewarded.rolling(window).mean(),color=sns.color_palette()[2])
        ax_correct_rolling.set_xlabel('session time (secs)')
        ax_correct_rolling.set_ylabel('% corr ('+str(window)+' trial average)')
        ax_correct_rolling.axhline(0.5,ls='--')
        ax_correct_rolling.set_ylim(0,1)

        # scatter of reaction times 
        sns.scatterplot(data=df, x='start_time',y="reaction_time",hue='target_image',style = 'rewarded', style_order=[True,False],
                    ax=ax_rt,legend=False)

        # histogram of reaction times 
        sns.histplot(data=df, x="reaction_time",hue='target_image',binrange=(0,12),bins=60,
                    ax=ax_rt_hist,legend=False,log_scale=log_hist)

        v = sns.violinplot(data=df, x='rewarded',y="reaction_time",hue='target_image',split = True, 
                    ax=ax_rt_violin)
        v.legend(bbox_to_anchor= (0.2, -0.2) );
        
        ax_rt_violin.set_yscale('log')
        ax_rt.set_yscale('log')
        if log_hist:ax_rt_hist.set_xscale('log')
        ax_rt_hist.set_xlim(0.1,12)

        fig.text(0.02,0.99, mouse, fontsize=14)
        fig.text(0.08,0.99, date, fontsize=14)
        fig.text(0.02,0.96, 'total percent correct: '+str(df[df.rewarded==True].shape[0]/float(df.shape[0])), fontsize=14)
        fig.text(0.02,0.93, 'median reaction time: '+str(np.nanmedian(df.reaction_time)), fontsize=14)
        # fig.text(0.02,0.90, 'mean rewarded time: '+str(np.nanmean(df[df.rewarded==True].hold_time)), fontsize=14)
    return fig

def generate_session_cheetah(mouse,date,return_=None,session='combine',log_hist=True):
    paths = glob.glob('/root/work/nwbs/'+mouse+'_'+date+'*_ce.nwb')
    if session == 'combine':
        paths = sort_day_sessions(paths);path=paths[0]
        io = NWBHDF5IO(path, mode='r')
        nwb_ = io.read()
        df = nwb_.trials.to_dataframe()
        for path in paths[1:]:
            df_ = NWBHDF5IO(path, mode='r').read().trials.to_dataframe()
            df = pd.concat([df,df_,],ignore_index=True)
    else:
        if len(paths) == 1:
            path = paths[0]
        if type(session)==int:
            path=paths[session]
        else: 
            print('incorrect format of session; provide either an int index to a sorted list of this days sessions, or str(combine) to combine all of this days sessions.')
            return np.nan
        df = NWBHDF5IO(path, mode='r').read().trials.to_dataframe()
    df['reaction_time']=df.response_time.values.astype(float) - df.start_time.values.astype(float)
    df['target_image']=df.image
    if return_ == 'df':
        return df

    if return_ == 'fig':
        fig = plt.figure(figsize=(8.5,11))
        ax_lever_all = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.0,1.0),yspan=(0,0.12))
        ax_lever_ex1 = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.0,0.2),yspan=(0.18,0.3))
        ax_lever_ex2 = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.26,0.46),yspan=(0.18,0.3))
        ax_lever_image = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.57,1.0),yspan=(0.2,0.35))
        ax_lever_response = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.57,1.0),yspan=(0.44,0.59))        
        ax_correct_rolling = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.,0.43),yspan=(0.39,0.59))
        ax_rt = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.,0.43),yspan=(0.6,0.8))
        ax_rt_hist = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.52,0.67),yspan=(0.66,0.8))
        ax_rt_violin = placeAxesOnGrid(fig,dim=(1,1),xspan=(0.78,1.0),yspan=(0.66,0.8))

        #plot lever - whole session
        if len(paths) == 1 or type(session)==int:
            nwb_ = NWBHDF5IO(path, mode='r').read()
            lever_timestamps = np.array(nwb_.acquisition['lever_position'].timestamps)
            lever_data = np.array(nwb_.acquisition['lever_position'].data)
        else:
            if session == 'combine':
                print('combining '+mouse+' sessions from '+date+'...')
                lever_timestamps = []
                lever_data =      []
                for path in paths:
                    nwb_ = NWBHDF5IO(path, mode='r').read()
                    lever_timestamps.extend(nwb_.acquisition['lever_position'].timestamps)
                    lever_data.extend(nwb_.acquisition['lever_position'].data)
                lever_timestamps = np.array(lever_timestamps)
                lever_data = np.array(lever_data)
            else:print('incorrect format of session; provide either an int index to a sorted list of this days sessions, or str(combine) to combine all of this days sessions.')
        ax_lever_all.plot(lever_timestamps,lever_data,'w')
        ax_lever_all.set_ylabel('lever position')
        ax_lever_all.set_xlabel('sesssion time (sec)')  
        
        # #plot some zoomed in lever examples
        r = np.random.randint(np.array(nwb_.acquisition['lever_position'].timestamps).shape[0])
        ax_lever_ex2.plot(nwb_.acquisition['lever_position'].timestamps[r:r+180],
                    nwb_.acquisition['lever_position'].data[r:r+180],color='w')
        r = np.random.randint(np.array(nwb_.acquisition['lever_position'].timestamps).shape[0])
        ax_lever_ex1.plot(nwb_.acquisition['lever_position'].timestamps[r:r+180],
                    nwb_.acquisition['lever_position'].data[r:r+180],color='w')

        # #plot lever around images A and B
        A_levers = np.zeros((df[df.target_image == 'elephant'].shape[0],180));al=0
        B_levers = np.zeros((df[df.target_image == 'cheetah'].shape[0],180));bl=0
        A_images = np.zeros((df[df.target_image == 'elephant'].shape[0],180));ai=0
        B_images = np.zeros((df[df.target_image == 'cheetah'].shape[0],180));bi=0
        for i,row in df.iterrows():
            lever_trigger = np.where((np.array(nwb_.acquisition['lever_position'].timestamps) > float(row.response_time)))[0][0]
            image_trigger = np.where((np.array(nwb_.acquisition['lever_position'].timestamps) > float(row.start_time)))[0][0]
            lever_response_d = nwb_.acquisition['lever_position'].data[lever_trigger-120:lever_trigger+60] 
            lever_response_t = nwb_.acquisition['lever_position'].timestamps[lever_trigger-120:lever_trigger+60] 
            lever_image_d = nwb_.acquisition['lever_position'].data[image_trigger-30:image_trigger+150] 
            lever_image_t = nwb_.acquisition['lever_position'].timestamps[image_trigger-30:image_trigger+150] 
            if row.target_image =='elephant':
                ax_lever_image.plot(np.linspace(-0.5,2.5,180),lever_image_d,lw=0.5,color=sns.color_palette()[0],alpha=0.3)
                A_levers[al,:]=lever_image_d; al+=1
            else:
                ax_lever_image.plot(np.linspace(-0.5,2.5,180),lever_image_d,lw=0.5,color=sns.color_palette()[1],alpha=0.3)
                B_levers[bl,:]=lever_image_d; bl+=1
            
            if row.target_image =='elephant':
                ax_lever_response.plot(np.linspace(-2,1,180),lever_response_d,lw=0.5,color=sns.color_palette()[0],alpha=0.3)
                A_images[ai,:]=lever_response_d; ai+=1
            else:
                ax_lever_response.plot(np.linspace(-2,1,180),lever_response_d,lw=0.5,color=sns.color_palette()[1],alpha=0.3)
                B_images[bi,:]=lever_response_d; bi+=1
            
        ax_lever_image.plot(np.linspace(-0.5,2.5,180),np.mean(A_levers,axis=0),lw=2,color=sns.color_palette()[0])
        ax_lever_image.plot(np.linspace(-0.5,2.5,180),np.mean(B_levers,axis=0),lw=2,color=sns.color_palette()[1])
        ax_lever_response.plot(np.linspace(-2,1,180),np.mean(A_images,axis=0),lw=2,color=sns.color_palette()[0])
        ax_lever_response.plot(np.linspace(-2,1,180),np.mean(B_images,axis=0),lw=2,color=sns.color_palette()[1])
            
        ax_lever_image.set_title('aligned to image onset')
        ax_lever_response.set_title('aligned lever movement')
        for ax in [ax_lever_image,ax_lever_response]:
            ax.set_xlabel('time (sec)',fontsize=10)
            ax.set_ylabel('lever (arb)',fontsize=10)

        window = 10
        ax_correct_rolling.plot(df[df.target_image=='elephant'].start_time,
        df[df.target_image=='elephant'].rewarded.rolling(window).mean(),color=sns.color_palette()[0])
        ax_correct_rolling.plot(df[df.target_image=='cheetah'].start_time,
        df[df.target_image=='cheetah'].rewarded.rolling(window).mean(),color=sns.color_palette()[1])
        ax_correct_rolling.plot(df.start_time,df.rewarded.rolling(window).mean(),color=sns.color_palette()[2])
        ax_correct_rolling.set_xlabel('session time (secs)')
        ax_correct_rolling.set_ylabel('% corr ('+str(window)+' trial average)')
        ax_correct_rolling.axhline(0.5,ls='--')
        ax_correct_rolling.set_ylim(0,1)

        # scatter of reaction times 
        sns.scatterplot(data=df, x='start_time',y="reaction_time",hue='target_image',style = 'rewarded', style_order=[True,False],
                    ax=ax_rt,legend=False)

        # histogram of reaction times 
        sns.histplot(data=df, x="reaction_time",hue='target_image',binrange=(0,12),bins=60,
                    ax=ax_rt_hist,legend=False,log_scale=log_hist)

        v = sns.violinplot(data=df, x='rewarded',y="reaction_time",hue='target_image',split = True, 
                    ax=ax_rt_violin)
        v.legend(bbox_to_anchor= (-3.5, -0.2) );
        
        ax_rt_violin.set_yscale('log')
        ax_rt.set_yscale('log')
        if log_hist:ax_rt_hist.set_xscale('log')
        ax_rt_hist.set_xlim(0.1,12)

    fig.text(0.02,0.99, mouse, fontsize=14)
    fig.text(0.08,0.99, date, fontsize=14)
    fig.text(0.02,0.96, 'total percent correct: '+str(df[df.rewarded==True].shape[0]/float(df.shape[0])), fontsize=14)
    fig.text(0.02,0.93, 'mean reaction time: '+str(np.nanmean(df.reaction_time)), fontsize=14)
    # fig.text(0.02,0.90, 'mean rewarded time: '+str(np.nanmean(df[df.rewarded==True].hold_time)), fontsize=14)
    return fig

def placeAxesOnGrid(fig,dim=[1,1],xspan=[0,1],yspan=[0,1],wspace=None,hspace=None):
    '''
    Takes a figure with a gridspec defined and places an array of sub-axes on a portion of the gridspec
    
    Takes as arguments:
        fig: figure handle - required
        dim: number of rows and columns in the subaxes - defaults to 1x1
        xspan: fraction of figure that the subaxes subtends in the x-direction (0 = left edge, 1 = right edge)
        yspan: fraction of figure that the subaxes subtends in the y-direction (0 = top edge, 1 = bottom edge)
        wspace and hspace: white space between subaxes in vertical and horizontal directions, respectively
        
    returns:
        subaxes handles
        
        written by doug ollerenshaw
    '''
    outer_grid = gridspec.GridSpec(100,100)
    inner_grid = gridspec.GridSpecFromSubplotSpec(dim[0],dim[1],
                                                  subplot_spec=outer_grid[int(100*yspan[0]):int(100*yspan[1]),int(100*xspan[0]):int(100*xspan[1])],
                                                  wspace=wspace, hspace=hspace)
    #NOTE: A cleaner way to do this is with list comprehension:
    # inner_ax = [[0 for ii in range(dim[1])] for ii in range(dim[0])]
    inner_ax = dim[0]*[dim[1]*[fig]] #filling the list with figure objects prevents an error when it they are later replaced by axis handles
    inner_ax = np.array(inner_ax)
    idx = 0
    for row in range(dim[0]):
        for col in range(dim[1]):
            inner_ax[row][col] = plt.Subplot(fig, inner_grid[idx])
            fig.add_subplot(inner_ax[row,col])
            idx += 1
    inner_ax = np.array(inner_ax).squeeze().tolist() #remove redundant dimension
    return inner_ax

def generate_session_figure(nwb_path):
    io = NWBHDF5IO(nwb_path, mode='r')
    nwb_ = io.read()

    #get task and phase information from the nwb
    if task == 'cheetah_or_elephant':
        pass
    if task == 'bonsai_levertask':
        generate_session(path)
    
    #choose which generate_  function to call
        
    # def generate_session(mouse,date,return_=None,session='combine'):
    #     # paths = glob.glob('/root/work/nwbs_lever/'+mouse+'_'+date+'*_lever.nwb')
    #     paths
    #     if len(paths)> 1:
    #         if session == 'combine':
    #             paths = sort_day_sessions(paths)
    #             df = combine_nwb_sessions(paths)
    #         else:
    #             nwb_path = paths[session]
    #             io = NWBHDF5IO(nwb_path, mode='r')
    #             nwb_ = io.read()
    #             df = nwb_.trials.to_dataframe()
    #     else: 
    #         nwb_path = paths[0]
    #         io = NWBHDF5IO(nwb_path, mode='r')
    #         nwb_ = io.read()
    #         df = nwb_.trials.to_dataframe()


    #     if return_ == 'df':
    #         return df

    #     if return_ == 'fig':

def get_history_sessions(today,tomonth,toyear,number_of_days=10):
    days = []
    mdays = [31,31,28,31,30,31,30,31,31,30,31,30]
    mmonths = [12,1,2,3,4,5,6,7,8,9,10,11]
    for day_minus in range(number_of_days):
        day_ = int(today)-day_minus
        if day_ < 1:
            last_month = int(tomonth)-1
            day_ = mdays[last_month]+day_
            if mmonths[last_month] > int(tomonth):
                lastyear = str(int(toyear) - 1)
                y_m_d = lastyear+'_'+str(mmonths[last_month])+'_'+str(day_)
            else: y_m_d = toyear+'_'+str(mmonths[last_month])+'_'+str(day_)
        else:
            y_m_d = toyear+'_'+tomonth+'_'+str(day_)
        days.extend([y_m_d])
    return days


import pandas as pd
import numpy as np
from dlab.generalephys import get_waveform_duration,get_waveform_PTratio,get_waveform_repolarizationslope,option234_positions
from scipy.cluster.vq import kmeans2
import seaborn as sns;sns.set_style("ticks")
import matplotlib.pyplot as plt
import h5py
import matplotlib.path as mpath
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
from matplotlib.collections import PatchCollection
import os

def get_peak_waveform_from_template(template):
    max = 0
    ind=0
    peak = np.zeros(np.shape(template.T)[0])
    for i,wv in enumerate(template.T):
        if np.max(np.abs(wv)) > max:
            max = np.max(np.abs(wv))
            ind = i
            peak = wv
    return peak

def df_from_phy(folder,expnum='1',recnum='1',site_positions = option234_positions,**kwargs):
    if 'est' not in folder:
        base_folder = os.path.basename(folder)
        cohort_ = os.path.basename(base_folder).split('_')[-2]
        mouse_  = os.path.basename(base_folder).split('_')[-1]

        #traverse down tree to data
        if 'open-ephys-neuropix' in base_folder:
            try:
                rec_folder = glob.glob(folder+'/*')[0]
            except:
                print(base_folder)
                return None
        else:
            rec_folder = folder
        raw_path = os.path.join(rec_folder,'experiment'+str(expnum),'recording'+str(recnum),'continuous')
        if len(glob.glob(raw_path+'/*100.0*'))>0:
            raw_path = glob.glob(raw_path+'/*100.0*')[0]
            print('loading from '+raw_path)
        else:
            
            print('could not find data folder for '+raw_path)
    
        if os.path.isfile(os.path.join(raw_path,'spike_clusters.npy')) :                  
    #             df = df_from_phy(raw_path,site_positions = ephys.option234_positions,cluster_file='KS2',cohort=cohort,mouse=mouse)

            path = raw_path
            units = ephys.load_phy_template(path,cluster_file='KS2',site_positions=site_positions)
            #structures is a dictionary that defines the bounds of the structure e.g.:{'v1':(0,850), 'hpc':(850,2000)}
            mouse = [];experiment=[];cell = [];ypos = [];xpos = [];waveform=[];template=[];structure=[];times=[]
            index = []; count = 1; cohort = []
            probe_id=[]
            depth=[];#print(list(nwb_data.keys()));print(list(nwb_data['processing'].keys()));

            for unit in list(units.keys()):
                if 'probe' in kwargs.keys():
                    probe_id.extend([kwargs['probe']])
                else:
                    probe_id.extend(['A'])
                if 'mouse' in kwargs.keys():
                    mouse.extend([kwargs['mouse']])
                else:
                    mouse.extend([mouse_])
                if 'experiment' in kwargs.keys():
                    experiment.extend([kwargs['experiment']])
                else:
                    experiment.extend(['placeholder'])
                if 'cohort' in kwargs.keys():
                    cohort.extend([kwargs['cohort']])
                else:
                    cohort.extend([cohort_])

                xpos.extend([units[unit]['xpos']])
                ypos.extend([units[unit]['ypos']])
                template.extend([units[unit]['template']])
                times.append(units[unit]['times'])
                waveform.append(units[unit]['waveform_weights'])

            df = pd.DataFrame(index=index)
            df = df.fillna(np.nan)
        #     df['nwb_id'] = nwb_id
            df['mouse'] = mouse
            df['experiment'] = experiment
            df['probe'] = probe_id
        #     df['structure'] = structure
            df['cell'] = units.keys()
            df['cohort'] = cohort
            df['times'] = times
            df['ypos'] = ypos
            df['xpos'] = xpos
    #         df['depth'] = depth
            df['waveform'] = waveform
            df['template'] = template
            return df

def df_from_nwb(nwb_data,structures=None,insertion_angle=55,nwbid=0):
    if type(nwb_data)==str:
        #print(nwb_data)
        nwbid = nwb_data
        nwb_data = h5py.File(nwb_data)
    else:
        nwb_data = nwb_data
    #structures is a dictionary that defines the bounds of the structure e.g.:{'v1':(0,850), 'hpc':(850,2000)}
    mouse = [];experiment=[];cell = [];ypos = [];xpos = [];waveform=[];template=[];structure=[];times=[]
    index = []; count = 1
    nwb_id = [];probe_id=[]
    depth=[];#print(list(nwb_data.keys()));print(list(nwb_data['processing'].keys()));
    if 'processing' in nwb_data.keys():
        for probe in list(nwb_data['processing'].keys()):
            if 'UnitTimes' in list(nwb_data['processing'][probe].keys()): 
                for i,u in enumerate(list(nwb_data['processing'][probe]['UnitTimes'].keys())):
                    if u != 'unit_list':
                        nwb_id.append(nwbid)
                        probe_id.append(probe)
                        index.append(count);count+=1
                        mouse.append(str(np.array(nwb_data.get('identifier'))))
                        experiment.append(1)
                        cell.append(u)
                        times.append(np.array(nwb_data['processing'][probe]['UnitTimes'][u]['times']));# print(list(nwb_data['processing'][probe]['UnitTimes'][u].keys()))
                        if 'ypos' in list(nwb_data['processing'][probe]['UnitTimes'][u].keys()):
                            ypos.append(np.array(nwb_data['processing'][probe]['UnitTimes'][u]['ypos']))
                            has_ypos = True
                        else:
                            ypos.append(None)
                            has_ypos = False				
                        if 'depth' in list(nwb_data['processing'][probe]['UnitTimes'][u].keys()):
                            depth.append(np.array(nwb_data['processing'][probe]['UnitTimes'][u]['depth']))
                        else:
                            if has_ypos:
                                depth.append(np.array(nwb_data['processing'][probe]['UnitTimes'][u]['ypos']))
                            else:
                                depth.append(None)
                        if 'xpos' in list(nwb_data['processing'][probe]['UnitTimes'][u].keys()):
                            xpos.append(np.array(nwb_data['processing'][probe]['UnitTimes'][u]['xpos']))
                            has_xpos = True
                        else:
                            xpos.append(None)
                            has_xpos = False
                        template.append(np.array(nwb_data['processing'][probe]['UnitTimes'][u]['template']))
                        waveform.append(get_peak_waveform_from_template(template[-1]))
                        if not structures == None:
                            structur = None
                            for struct, bounds in structures.iteritems():
                                if ypos[-1] > bounds[0] and ypos[-1]< bounds[1] :
                                    structur=struct
                        else:
                            structur = None
                        structure.append(structur)
    df = pd.DataFrame(index=index)
    df = df.fillna(np.nan)
    df['nwb_id'] = nwb_id
    df['mouse'] = mouse
    df['experiment'] = experiment
    df['probe'] = probe_id
    df['structure'] = structure
    df['cell'] = cell
    df['times'] = times
    df['ypos'] = ypos
    df['xpos'] = xpos
    df['depth'] = depth
    df['waveform'] = waveform
    df['template'] = template
    return df

def classify_waveform_shape(df,plots=False,save_plots=False,basepath='',kmeans=0):
	durations = np.zeros(np.shape(df)[0])
	PTratio = np.zeros(np.shape(df)[0])
	repolarizationslope = np.zeros(np.shape(df)[0])
	for i,waveform in enumerate(df.waveform):
        # try:
		durations[i]=get_waveform_duration(waveform)
		PTratio[i]=get_waveform_PTratio(waveform)
		repolarizationslope[i]=get_waveform_repolarizationslope(waveform,window=18)
        # except:
        #     durations[i]=np.nan
        #     PTratio[i]=np.nan
        #     repolarizationslope[i]=np.nan
	df['waveform_duration'] = durations
	df['waveform_PTratio'] = PTratio
	df['waveform_repolarizationslope'] = repolarizationslope

	waveform_k = kmeans2(np.vstack(((durations-np.min(durations))/np.max((durations-np.min(durations))),
									(PTratio-np.min(PTratio))/np.max((PTratio-np.min(PTratio))),
									(repolarizationslope-np.min(repolarizationslope))/np.max((repolarizationslope-np.min(repolarizationslope))))).T,
							2, iter=300, thresh=5e-6,minit='points')
    # waveform_k = kmeans2(np.vstack((durations/np.max(durations),PTratio/np.max(PTratio))).T, 2, iter=300, thresh=5e-6,minit='points')
    # waveform_k = kmeans2(np.vstack((durations/np.max(durations),(repolarizationslope-np.min(repolarizationslope))/np.max(repolarizationslope))).T, 2, iter=900, thresh=5e-7,minit='points')
    
    #assign fs and rs to the kmeans results
	if np.mean(durations[np.where(waveform_k[1]==0)[0]]) < np.mean(durations[np.where(waveform_k[1]==1)[0]]):
		fs_k = 0;rs_k = 1
		waveform_class_ids = ['fs','rs']
	else:
		rs_k = 0;fs_k = 1
		waveform_class_ids = ['rs','fs']
	waveform_class = [waveform_class_ids[k] for k in waveform_k[1]]

	#uncomment this to ignore the preceding kmeans and just split on the marginal distribution of durations
	if kmeans==0:
		waveform_class = ['fs' if duration < 0.0004 else 'rs' for i,duration in enumerate(durations) ]
	else:
		waveform_k = kmeans2(np.vstack(((durations-np.min(durations))/np.max((durations-np.min(durations))),
										(PTratio-np.min(PTratio))/np.max((PTratio-np.min(PTratio))),
										(repolarizationslope-np.min(repolarizationslope))/np.max((repolarizationslope-np.min(repolarizationslope))))).T,
								kmeans, iter=300, thresh=5e-6,minit='points')
		# waveform_k = kmeans2(np.vstack((durations/np.max(durations),PTratio/np.max(PTratio))).T, 2, iter=300, thresh=5e-6,minit='points')
		# waveform_k = kmeans2(np.vstack((durations/np.max(durations),(repolarizationslope-np.min(repolarizationslope))/np.max(repolarizationslope))).T, 2, iter=900, thresh=5e-7,minit='points')
		
		#assign fs and rs to the kmeans results
		if np.mean(durations[np.where(waveform_k[1]==0)[0]]) < np.mean(durations[np.where(waveform_k[1]==1)[0]]):
			fs_k = 0;rs_k = 1
			waveform_class_ids = ['fs','rs']
		else:
			rs_k = 0;fs_k = 1
			waveform_class_ids = ['rs','fs']
		waveform_class = [waveform_class_ids[k] for k in waveform_k[1]]

	#force upwards spikes to have the own class, because we're not sure how they fit in this framework
	waveform_class = [waveform_class[i] if ratio < 1.0 else 'up' for i,ratio in enumerate(PTratio) ]
	df['waveform_class']=waveform_class

	#mark narrow upwards spikes as axons
	waveform_class = ['axon' if all([duration < 0.0004,waveform_class[i]=='up']) else waveform_class[i] for i,duration in enumerate(durations) ]
	df['waveform_class']=waveform_class

	# #mark narrow downward spike at the very bottom of cortex as axons
	#waveform_class = ['axon' if all([duration < 0.0004,waveform_class[i]=='fs',df['depth'][i+1] > 750, df['depth'][i+1]<1050]) else waveform_class[i] for i,duration in enumerate(durations) ]
	df['waveform_class']=waveform_class

	if plots:
		plot_waveform_classification(durations, PTratio, repolarizationslope,df,save_plots=save_plots,basepath=basepath)
	return df

def plot_waveform_classification(durations, PTratio, repolarizationslope, df,save_plots=False, basepath=''):
	f,ax = plt.subplots(1,3,figsize=(8,3))
	ax[0].plot(durations[np.where(df.waveform_class=='rs')[0]],PTratio[np.where(df.waveform_class=='rs')[0]],'o',ms=3.2)
	ax[0].plot(durations[np.where(df.waveform_class=='fs')[0]],PTratio[np.where(df.waveform_class=='fs')[0]],'o',ms=3.2)
	#ax[0].plot(durations[np.where(df.waveform_class=='up')[0]],PTratio[np.where(df.waveform_class=='up')[0]],'o',ms=3.2)
	ax[0].plot(durations[np.where(df.waveform_class=='axon')[0]],PTratio[np.where(df.waveform_class=='axon')[0]],'o',ms=3.2)
	ax[0].set_xlabel('width (sec)')
	ax[0].set_ylabel('peak/trough ratio')
	ax[1].plot(durations[np.where(df.waveform_class=='rs')[0]],repolarizationslope[np.where(df.waveform_class=='rs')[0]],'o',ms=3.2)
	ax[1].plot(durations[np.where(df.waveform_class=='fs')[0]],repolarizationslope[np.where(df.waveform_class=='fs')[0]],'o',ms=3.2)
	#ax[1].plot(durations[np.where(df.waveform_class=='up')[0]],repolarizationslope[np.where(df.waveform_class=='up')[0]],'o',ms=3.2)
	ax[1].plot(durations[np.where(df.waveform_class=='axon')[0]],repolarizationslope[np.where(df.waveform_class=='axon')[0]],'o',ms=3.2)
	ax[1].set_xlabel('width (sec)')
	ax[1].set_ylabel('repolarization slope')
	ax[2].plot(PTratio[np.where(df.waveform_class=='rs')[0]],repolarizationslope[np.where(df.waveform_class=='rs')[0]],'o',ms=3.2)
	ax[2].plot(PTratio[np.where(df.waveform_class=='fs')[0]],repolarizationslope[np.where(df.waveform_class=='fs')[0]],'o',ms=3.2)
	#ax[2].plot(PTratio[np.where(df.waveform_class=='up')[0]],repolarizationslope[np.where(df.waveform_class=='up')[0]],'o',ms=3.2)
	ax[2].plot(PTratio[np.where(df.waveform_class=='axon')[0]],repolarizationslope[np.where(df.waveform_class=='axon')[0]],'o',ms=3.2)
	ax[2].set_ylabel('repolarization slope')
	ax[2].set_xlabel('peak/trough ratio')
	ax[0].set_xlim(0.0,0.0015);ax[1].set_xlim(0.0,0.0015)
	ax[0].set_ylim(0,1.1);ax[2].set_xlim(0,1.1)
	plt.tight_layout()
	for axis in ax:
	#    ephys.cleanAxes(axis,bottomLabels=True,leftLabels=True)
		axis.locator_params(axis='x',nbins=4)
	ax[2].legend(loc='upper right')
	panelname = 'waveforms_clusters'
	plt.tight_layout()
	if save_plots:
		plt.gcf().savefig(os.path.join(basepath,'figures','panels',panelname+'.png'))
		plt.gcf().savefig(os.path.join(basepath,'figures','panels',panelname+'.eps'))
		
	nbins = 36
	plt.hist(durations[np.where(df.waveform_class=='rs')[0]],range=(0,0.0015),bins=nbins)
	plt.hist(durations[np.where(df.waveform_class=='fs')[0]],range=(0,0.0015),bins=nbins)
	plt.hist(durations[np.where(df.waveform_class=='axon')[0]],range=(0,0.0015),bins=nbins)
	plt.figure()
	plt.hist((durations[np.where(df.waveform_class=='rs')[0]],durations[np.where(df.waveform_class=='fs')[0]],durations[np.where(df.waveform_class=='axon')[0]]),range=(0,0.0015),bins=nbins,stacked=True)
	#ephys.cleanAxes(plt.gca(),bottomLabels=True,leftLabels=True)
	plt.xlabel('waveform duration (sec)')
	plt.ylabel('neuron count')
	panelname = 'waveforms_durationhistogram'
	plt.tight_layout()
	if save_plots:
		plt.gcf().savefig(os.path.join(basepath,'figures','panels',panelname+'.png'))
		plt.gcf().savefig(os.path.join(basepath,'figures','panels',panelname+'.eps'))
	
	plt.figure(figsize=(4,3))
	
	waveform_time = np.linspace(-1*np.where(df.waveform[1] > 0.)[0][0]/30000.,(len(df.waveform[1])-np.where(df.waveform[1] > 0.)[0][0])/30000.,len(df.waveform[1]))*1000
	#plot all
	for i,waveform in enumerate(df.waveform):
		#waveform_time = np.linspace(0,len(waveform)/30000.,len(waveform))*1000
		if df.waveform_class[i]=='rs':
			plt.plot(waveform_time,waveform/np.max(np.abs(waveform)),color=sns.color_palette()[0],alpha=0.01,ms=0.5)
		if df.waveform_class[i]=='axon':#df.waveform_class.unique()[np.where(df.waveform_class=='axon')[0]]:
			plt.plot(waveform_time,waveform/np.max(np.abs(waveform)),color=sns.color_palette()[2],alpha=0.01,ms=0.5)
		if df.waveform_class[i]=='fs':#df.waveform_class.unique()[np.where(df.waveform_class=='fs')[0]]:
				plt.plot(waveform_time,waveform[:len(waveform_time)]/np.max(np.abs(waveform[:len(waveform_time)])),color=sns.color_palette()[1],alpha=0.01,ms=0.5)
	# plot means, normalized
	for waveform_class in ['rs','fs','axon']:#df.waveform_class.unique():
		if waveform_class != 'up':# and waveform_class!='axon':
			a = np.mean(df.waveform[df.waveform_class==waveform_class])
			plt.plot(waveform_time,a[:len(waveform_time)]\
							/(np.max(np.abs(np.mean(df.waveform[df.waveform_class==waveform_class])))),lw=4)
		#plt.plot(waveform_time,np.mean(df.waveform[df.waveform_class==waveform_class])/(np.max(np.abs(np.mean(df.waveform[df.waveform_class==waveform_class])))),lw=2)
	# plt.plot(waveform_time,np.mean(df.waveform[df.waveform_class=='rs'])/(np.min(np.mean(df.waveform[df.waveform_class=='rs']))*-1),lw=2)
	# plt.plot(waveform_time,np.mean(df.waveform[df.waveform_class=='up'])/(np.max(np.mean(df.waveform[df.waveform_class=='up']))),lw=2)
	plt.title('RS: '+str(len(df.waveform_class[df.waveform_class=='rs']))+
				'   FS: '+str(len(df.waveform_class[df.waveform_class=='fs']))+
			  '   axon: '+str(len(df.waveform_class[df.waveform_class=='axon'])))#+
	#            '   up:'+str(len(df.waveform_class[df.waveform_class=='up'])))
	
	
	plt.gca().set_xlim(0.,2.0)
	plt.gca().legend(loc='upper left')
	#ephys.cleanAxes(plt.gca(),leftLabels=True,bottomLabels=True)
	plt.gca().set_ylabel('normalized amplitude',size=10)
	d=plt.gca().set_xlabel('time (msec)',size=10)
	panelname = 'waveforms_mean_peak'
	plt.tight_layout()
	if save_plots:
		plt.gcf().savefig(os.path.join(basepath,'figures','panels',panelname+'.png'))
		plt.gcf().savefig(os.path.join(basepath,'figures','panels',panelname+'.eps'))

def drawPhaseIIIProbe(colors,ax=-1,highlight=-1,clim=None, cmap='viridis', drawLines=False):
	'''
	Args:
		colors: a list of values to plotted as colors on the probe
		ax
		highlight
		clim: color map limits
		cmap: color map to use; default viridis
		drawLines: whether or not to draw the outline of the probe; default is False
	Returns:
		None, plots an image of the input colors on a Phase3A Neuropixels probes
	written by josh siegle
	'''
	if ax == -1:
		fig, ax = plt.subplots()
	
	patches = []
	
	for ch in range(0,len(colors)):
		
		channelPos = ch % 4
		channelHeight = ch / 4
		
		if channelPos == 0:
			xloc = -1.5
			yloc = channelHeight*2
		elif channelPos == 1:
			xloc = 0.5
			yloc = channelHeight*2
		elif channelPos == 2:
			xloc = -0.5
			yloc = channelHeight*2 + 1
		else:
			xloc = 1.5
			yloc = channelHeight*2 + 1
	
		rect = mpatches.Rectangle([xloc, yloc], 1.0, 2.0, ec="none", ls='None')
		
		if drawLines:
			if ch % 50 == 0:
				plt.plot([-5, 6], [yloc, yloc], 'gray')
				
			if ch % 100 == 0:
				plt.plot([-5, 6], [yloc, yloc], '-k')
			
		patches.append(rect)
		
		if ch == highlight:
			highlightX = xloc
			highlightY = yloc
			highlight = 1
		
	collection = PatchCollection(patches, cmap=cmap)
	
	collection.set_array(colors)
	if clim != None:
		collection.set_clim(clim[0],clim[1])
	ax.add_collection(collection)
	
	for ch in np.arange(0,len(colors),50):
		plt.plot([-2.5,-2],[ch/2, ch/2],'k')
	
	
	if highlight > -1:
		print(highlightY)
		plt.plot(highlightX, highlightY, color=[1,1,1])
	
	plt.axis('off')
	plt.xlim((-5,6))
	plt.ylim((-5,ch/2 + 20))

def get_spike_limits(nwb_data):
    firsts = [np.array(nwb_data['processing'][nwb_data['processing'].keys()[0]]['UnitTimes'][other]['times'])[0]\
              for other in np.array(nwb_data['processing'][nwb_data['processing'].keys()[0]]['UnitTimes']['unit_list'])]
    lasts = [np.array(nwb_data['processing'][nwb_data['processing'].keys()[0]]['UnitTimes'][other]['times'])[-1]\
              for other in np.array(nwb_data['processing'][nwb_data['processing'].keys()[0]]['UnitTimes']['unit_list'])]
    return np.min(firsts),np.max(lasts)

def cleanAxes(ax,bottomLabels=False,leftLabels=False,rightLabels=False,topLabels=False,total=False):
    ax.tick_params(axis='both',labelsize=10)
    ax.spines['top'].set_visible(False);
    ax.yaxis.set_ticks_position('left');
    ax.spines['right'].set_visible(False);
    ax.xaxis.set_ticks_position('bottom')
    if not bottomLabels or topLabels:
        ax.set_xticklabels([])
    if not leftLabels or rightLabels:
        ax.set_yticklabels([])
    if rightLabels:
        ax.spines['right'].set_visible(True);
        ax.spines['left'].set_visible(False);
        ax.yaxis.set_ticks_position('right');
    if total:
        ax.set_frame_on(False);
        ax.set_xticklabels('',visible=False);
        ax.set_xticks([]);
        ax.set_yticklabels('',visible=False);
        ax.set_yticks([])